import torch
import torch.nn as nn
from typing import Optional, Union, Iterable, Dict, List
import pythonbasictools as pybt
from . import BaseRegularization
from ..transforms.base import to_numpy
from ..init import dale_


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[BaseRegularization])
class DaleLawL2(BaseRegularization):
    """
    Regularisation of the connectome to apply Dale's law and L2. In a nutshell, the Dale's law
    stipulate that neurons can either have excitatory or inhibitory connections, not both.
    The L2 regularisation reduce the energy of the network. This regularisation allow you to follow the
    Dale's law and/or L2 depending on the factor alpha. The equation used is showed by :eq:`dale_l2`.

    .. math::
        :label: dale_l2

        \\begin{equation}
            \\mathcal{L}_{\\text{DaleLawL2}} = \\text{Tr}\\left( W^T \\left(\\alpha W - \\left(1 - \\alpha\\right) W_{\\text{ref}}\\right) \\right)
        \\end{equation}


    In the case where :math:`\\alpha = 0`, the regularisation will only follow the Dale's law shown by :eq:`dale`.

    .. math::
        :label: dale

        \\begin{equation}
            \\mathcal{L}_{\\text{DaleLaw}} = -\\text{Tr}\\left( W^T W_{\\text{ref}}\\right)
        \\end{equation}

    In the case where :math:`\\alpha = 1`, the regularisation will only follow the L2 regularisation shown by :eq:`l2`.

    .. math::
        :label: l2

        \\begin{equation}
            \\mathcal{L}_{\\text{L2}} = \\text{Tr}\\left( W^T W\\right)
        \\end{equation}


    :Attributes:
        - :attr:`alpha` (float): Number between 0 and 1 that favors one of the constraints.
        - :attr:`dale_kwargs` (dict): kwargs of the Dale's law. See :func:`dale_`.
        - :attr:`reference_weights` (Iterable[torch.Tensor]): Reference weights to compare. Must be the same size as the weights.

    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            alpha: float = 0.8,
            reference_weights: Optional[Iterable[torch.Tensor]] = None,
            Lambda: float = 1.0,
            optimizer: Optional[torch.optim.Optimizer] = None,
            **dale_kwargs
    ):
        """
        :param params: Weights matrix to regularize (can be multiple)
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param alpha: Number between 0 and 1 that favors one of the constraints.
            If alpha = 0 -> Only Dale's law is applied.
            If alpha = 1 -> Only the reduction of the energy is applied.
            If 1 < alpha < 0 -> Both Dale's law and the reduction of the energy are applied with their ratio.
        :type alpha: float
        :param reference_weights: Reference weights to compare. Must be the same size as the weights. If not provided,
            the weights will be generated automatically with the dale_kwargs.
        :type reference_weights: Optional[Iterable[torch.Tensor]]
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float
        :param dale_kwargs: kwargs of the Dale's law.

        :keyword float inh_ratio: ratio of inhibitory connections. Must be between 0 and 1.
        :keyword float rho: The connectivity ratio. Must be between 0 and 1. If rho = 1, the tensor will be fully connected.
        :keyword bool inh_first: If True, the inhibitory neurons will be in the first half of the tensor. If False,
            the neurons will be shuffled.
        :keyword Optional[int] seed: seed for the random number generator. If None, the seed is not set.
        """
        super(DaleLawL2, self).__init__(params, Lambda, optimizer=optimizer)
        self.__name__ = self.__class__.__name__
        self.alpha = alpha
        if self.alpha > 1 or self.alpha < 0:
            raise ValueError("alpha must be between 0 and 1")
        self.dale_kwargs = dale_kwargs
        self.reference_weights = self._init_reference_weights(reference_weights)

    def _init_reference_weights(
            self,
            reference_weights: Optional[Union[Iterable[torch.Tensor], Dict[str, torch.Tensor]]] = None
    ):
        """
        Initialize the reference weights with Dale's law.
        """
        if reference_weights is None:
            self.reference_weights = []
            for param in self.params:
                self.reference_weights.append(torch.sign(dale_(torch.empty_like(param), **self.dale_kwargs)))
        else:
            self.reference_weights = [torch.sign(ref) for ref in reference_weights]
        return self.reference_weights

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        Compute the forward pass of the Dale's law. If alpha = 1 and the reference weights is not provided, it will be
        modified to 0, so it can get cancel.

        :param args: weights matrix
        :param kwargs: kwargs of the forward pass
        """
        loss_list = []
        for param, ref in zip(self.params, self.reference_weights):
            loss = torch.trace(
                param.T @ (self.alpha * param - (1 - self.alpha) * ref.to(param.device))
            )
            loss_list.append(loss)
        if len(self.params) == 0:
            loss = torch.tensor(0.0, dtype=torch.float32)
        else:
            loss = torch.sum(torch.stack(loss_list))
        return loss


class DaleLaw(DaleLawL2):
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            reference_weights: Optional[Iterable[torch.Tensor]] = None,
            Lambda: float = 1.0,
            **dale_kwargs
    ):
        """
        :param params: Weights matrix to regularize (can be multiple)
        :param reference_weights: Reference weights to compare. Must be the same size as the weights. If not provided,
            the weights will be generated automatically with the dale_kwargs.
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :param dale_kwargs: kwargs of the Dale's law.

        :keyword inh_ratio: ratio of inhibitory connections. Must be between 0 and 1.
        :keyword rho: The connectivity ratio. Must be between 0 and 1. If rho = 1, the tensor will be fully connected.
        :keyword inh_first: If True, the inhibitory neurons will be in the first half of the tensor. If False,
            the neurons will be shuffled.
        :keyword seed: seed for the random number generator. If None, the seed is not set.
        """
        super(DaleLaw, self).__init__(params, 0.0, reference_weights, Lambda, **dale_kwargs)
        self.__name__ = self.__class__.__name__


class WeightsDistance(BaseRegularization):
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            reference_weights: Iterable[torch.Tensor],
            Lambda: float = 1.0,
            p: int = 1,
    ):
        super(WeightsDistance, self).__init__(params, Lambda)
        self.reference_weights = list(reference_weights)
        self.p = p

    def forward(self, *args, **kwargs) -> torch.Tensor:
        loss_list = []
        for param, ref in zip(self.params, self.reference_weights):
            loss_list.append(torch.linalg.norm(torch.abs(param - ref), self.p))
        if len(self.params) == 0:
            loss = torch.tensor(0.0, dtype=torch.float32)
        else:
            loss = torch.sum(torch.stack(loss_list))
        return loss


class ExcRatioTargetRegularization(BaseRegularization):
    r"""
    Applies the function:

    .. math::
        \text{loss}(x) = \lambda \cdot \sum_{i=1}^N \left|(\text{mean}(\text{sign}(x_i)) + 1) - 2\cdot\text{target} \right|

    Where :math:`x` is the list of input parameters, :math:`N` is the number of parameter, :math:`\text{sign}(x_i)` is the
    sign of the element :math:`x_i`, :math:`\text{mean}(\text{sign}(x_i))` is the mean of the signs of the elements in the
    tensor, :math:`\text{target}` is the target value and :math:`\lambda` is the weight of the regularization.


    Examples::
        >>> import neurotorch as nt
        >>> layer = nt.WilsonCowanLayer(10, 10, force_dale_law=True)
        >>> m = ExcRatioTargetRegularization(params=layer.get_sign_parameters(), Lambda=0.1, exc_target_ratio=0.9)
        >>> loss = m()
    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            exc_target_ratio: float = 0.8,
            Lambda: float = 1.0,
            **kwargs
    ):
        """
        Create a new ExcRatioTargetRegularization.

        :param params: Weights matrix to regularize.
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param exc_target_ratio: Target ratio of excitatory neurons. Must be between 0 and 1.
        :type exc_target_ratio: float
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float

        :keyword kwargs: kwargs of the BaseRegularization.
        """
        super(ExcRatioTargetRegularization, self).__init__(params, Lambda, **kwargs)
        assert 0 < exc_target_ratio < 1, "exec_target_ratio must be between 0 and 1"
        self.exc_target_ratio = exc_target_ratio
        self.sign_func = kwargs.get("sign_func", torch.nn.Softsign())

    def forward(self, *args, **kwargs) -> torch.Tensor:
        loss_list = []
        for param in self.params:
            param_ratio = torch.mean(self.sign_func(param)) + 1
            loss_list.append(torch.abs(param_ratio - 2 * self.exc_target_ratio))
        if len(self.params) == 0:
            loss = torch.tensor(0.0, dtype=torch.float32, requires_grad=True)
        else:
            loss = torch.sum(torch.stack(loss_list))
        return loss

    def get_params_exc_ratio(self) -> List[float]:
        """
        Returns the excitatory ratio of each parameter.
        """
        return [to_numpy(((torch.mean(self.sign_func(param)) + 1)/2).item()) for param in self.params]

    def get_params_inh_ratio(self) -> List[float]:
        """
        Returns the inhibitory ratio of each parameter.
        """
        return [to_numpy(((1 - torch.mean(self.sign_func(param)))/2).item()) for param in self.params]

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        loss = to_numpy(self().item())
        exc_ratio = self.get_params_exc_ratio()
        return {"exc_ratio": exc_ratio, "exc_ratio_loss": loss}


class InhRatioTargetRegularization(ExcRatioTargetRegularization):
    """
    Applies the `ExcRatioTargetRegularization` with the target ratio of inhibitory neurons.
    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            inh_target_ratio: float = 0.2,
            Lambda: float = 1.0,
    ):
        """
        Create a new InhRatioTargetRegularization.

        :param params: Weights matrix to regularize.
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param inh_target_ratio: Target ratio of inhibitory neurons. Must be between 0 and 1.
        :type inh_target_ratio: float
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float
        """
        assert 0 < inh_target_ratio < 1, "inh_target_ratio must be between 0 and 1"
        super(InhRatioTargetRegularization, self).__init__(
            params=params,
            Lambda=Lambda,
            exc_target_ratio=1 - inh_target_ratio,
        )

    def on_pbar_update(self, trainer, **kwargs) -> dict:
        loss = to_numpy(self().item())
        inh_ratio = self.get_params_inh_ratio()
        return {"inh_ratio": inh_ratio, "inh_ratio_loss": loss}





import warnings
from collections import defaultdict
from typing import Optional, Sequence, Union, Dict, Callable, List, Mapping, Any

import torch

from .learning_algorithm import LearningAlgorithm
from ..transforms.base import to_numpy, to_tensor
from ..learning_algorithms.tbptt import TBPTT
from ..utils import (
    list_insert_replace_at,
    zero_grad_params,
    unpack_out_hh,
    recursive_detach,
    dy_dw_local,
    clip_tensors_norm_,
    recursive_detach_
)
from ..utils.random import unitary_rn_normal_matrix


class Eprop(TBPTT):
    r"""
    Apply the eligibility trace forward propagation (e-prop) :cite:t:`bellec_solution_2020`
    algorithm to the given model.


    .. image:: ../../images/learning_algorithms/EpropDiagram.png
        :width: 300
        :align: center


    Note: If this learning algorithm is used for classification, the output layer should have a log-softmax activation
        function and the target should be a one-hot encoded tensor. Then, the loss function should be the negative log
        likelihood loss function from :class:`nt.losses.NLLLoss` with the ``target_as_one_hot`` argument set to ``True``.

    """
    CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
    CHECKPOINT_FEEDBACK_WEIGHTS_KEY: str = "feedback_weights"
    OPTIMIZER_PARAMS_GROUP_IDX = 0
    OPTIMIZER_OUTPUT_PARAMS_GROUP_IDX = 1
    DEFAULT_OPTIMIZER_CLS = torch.optim.AdamW
    DEFAULT_Y_KEY = "default_key"
    DEFAULT_FEEDBACKS_GEN_STRATEGY = "xavier_normal"
    FEEDBACKS_GEN_FUNCS = {
        "randn": lambda *args, **kwargs: torch.randn(*args, **kwargs),
        "xavier_normal": lambda *args, **kwargs: torch.nn.init.xavier_normal_(
            torch.empty(*args), gain=kwargs.get("gain", 1.0)
        ),
        "kaiming_normal": lambda *args, **kwargs: torch.nn.init.kaiming_normal_(
            torch.empty(*args), a=kwargs.get("a", 0.0), mode=kwargs.get("mode", "fan_in"),
            nonlinearity=kwargs.get("nonlinearity", "leaky_relu"),
        ),
        "rand": lambda *args, **kwargs: torch.rand(*args, **kwargs),
        "ones": lambda *args, **kwargs: torch.ones(*args),
        "unitary": lambda *args, **kwargs: unitary_rn_normal_matrix(*args, **kwargs),
        "orthogonal": lambda *args, **kwargs: torch.nn.init.orthogonal_(
            torch.empty(*args), gain=kwargs.get("gain", 1.0)
        ),
    }
    DEFAULT_FEEDBACKS_STR_NORM_CLIP_VALUE = {
        "randn": 1.0,
        "xavier_normal": torch.inf,
        "kaiming_normal": torch.inf,
        "rand": 1.0,
        "ones": 1.0,
        "unitary": torch.inf,
        "orthogonal": torch.inf,
    }

    def __init__(
            self,
            *,
            params: Optional[Sequence[torch.nn.Parameter]] = None,
            output_params: Optional[Sequence[torch.nn.Parameter]] = None,
            layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
            output_layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
            **kwargs
    ):
        """
        Constructor for Eprop class.

        :param params: The hidden parameters to optimize. If not provided, eprop will try to find the hidden parameters
                        by looking for the parameters of the layers provided or in the inputs and hidden layers of the
                        model provided in the trainer.
        :type params: Optional[Sequence[torch.nn.Parameter]]
        :param output_params: The output parameters to optimize. If not provided, eprop will try to find the output
                        parameters by looking for the parameters of the output layers provided or in the output layers
                        of the model provided in the trainer.
        :type output_params: Optional[Sequence[torch.nn.Parameter]]
        :param layers: The hidden layers to optimize. If not provided, eprop will try to find the hidden layers
                        by looking for the layers of the model provided in the trainer.
        :type layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]]
        :param output_layers: The output layers to optimize. If not provided, eprop will try to find the output layers
                        by looking for the layers of the model provided in the trainer.
        :type output_layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]]
        :param kwargs: Keyword arguments to pass to the parent class and to configure the algorithm.

        :keyword Optional[torch.optim.Optimizer] optimizer: The optimizer to use. If provided make sure to provide the
                        param_group in the following format:
                                [{"params": params, "lr": params_lr}, {"params": output_params, "lr": output_params_lr}]
                        The index of the group must be the same as the OPTIMIZER_PARAMS_GROUP_IDX and
                        OPTIMIZER_OUTPUT_PARAMS_GROUP_IDX constants which are 0 and 1 respectively.
                        If not provided, torch.optim.Adam is used.
        :keyword Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] criterion: The
                        criterion to use for the output learning signal. If not provided, torch.nn.MSELoss is used. Note
                        that this criterion will be minimized.
        :keyword float params_lr: The learning rate for the hidden parameters. Defaults to 1e-4.
        :keyword float output_params_lr: The learning rate for the output parameters. Defaults to 2e-4.
        :keyword float eligibility_traces_norm_clip_value: The value to clip the eligibility traces norm to.
                        Defaults to torch.inf.
        :keyword float grad_norm_clip_value: The value to clip the gradients norm to. This parameter is used to
                        normalize the gradients of the parameters in order to help the convergence and avoid
                        overflowing. Defaults to 1.0.
        :keyword str feedbacks_gen_strategy: The strategy to use to generate the feedbacks. Defaults to
                        Eprop.DEFAULT_FEEDBACKS_GEN_STRATEGY which is "xavier_normal". The available strategies are
                        stored in Eprop.FEEDBACKS_GEN_FUNCS which are:
                            - "randn": Normal distribution with mean 0 and variance 1.
                            - "xavier_normal": Xavier normal distribution.
                            - "rand": Uniform distribution between 0 and 1.
                            - "unitary": Unitary matrix with normal distribution.
        :keyword float nan: The value to use to replace the NaN values in the gradients. Defaults to 0.0.
        :keyword float posinf: The value to use to replace the inf values in the gradients. Defaults to 1.0.
        :keyword float neginf: The value to use to replace the -inf values in the gradients. Defaults to -1.0.
        :keyword bool save_state: Whether to save the state of the optimizer. Defaults to True.
        :keyword bool load_state: Whether to load the state of the optimizer. Defaults to True.
        :keyword bool raise_non_finite_errors: Whether to raise non-finite errors when detected. Defaults to False.
        """
        kwargs.setdefault("save_state", True)
        kwargs.setdefault("load_state", True)
        kwargs.setdefault("backward_time_steps", 1)
        kwargs.setdefault("optim_time_steps", 1)
        kwargs.setdefault("criterion", torch.nn.MSELoss())
        kwargs.setdefault("alpha", 0.001)
        super().__init__(
            params=params,
            layers=layers,
            output_layers=output_layers,
            **kwargs
        )
        self.output_params = output_params or []
        self._feedbacks_gen_strategy = kwargs.get("feedbacks_gen_strategy", self.DEFAULT_FEEDBACKS_GEN_STRATEGY).lower()
        if self._feedbacks_gen_strategy not in self.FEEDBACKS_GEN_FUNCS:
            raise NotImplementedError(
                f"Feedbacks generation strategy '{self._feedbacks_gen_strategy}' is not implemented."
                f"Maybe you meant one of the following: {', '.join(self.FEEDBACKS_GEN_FUNCS.keys())}"
                f" or you can implement your own by adding it to the Eprop.FEEDBACKS_GEN_FUNCS dictionary."
                f" If your new strategy is a common one, please consider contributing to the library by opening a "
                f"pull request on https://github.com/NeuroTorch/NeuroTorch."
            )
        self.feedback_weights = None
        self.rn_gen = torch.Generator()
        self.rn_gen.manual_seed(kwargs.get("seed", 0))
        self.eligibility_traces = [torch.zeros_like(p) for p in self.params]
        self.output_eligibility_traces = [torch.zeros_like(p) for p in self.output_params]
        self.learning_signals = defaultdict(list)
        self.param_groups = []
        self.eval_criterion = kwargs.get("eval_criterion", self.criterion)
        self.gamma = kwargs.get("gamma", 0.001)
        self._default_params_lr = kwargs.get("params_lr", 1e-5)
        self._default_output_params_lr = kwargs.get("output_params_lr", 2e-5)
        self.eligibility_traces_norm_clip_value = to_tensor(kwargs.get("eligibility_traces_norm_clip_value", torch.inf))
        self.learning_signal_norm_clip_value = to_tensor(kwargs.get("learning_signal_norm_clip_value", torch.inf))
        self.feedback_weights_norm_clip_value = to_tensor(kwargs.get(
            "feedback_weights_norm_clip_value",
            self.DEFAULT_FEEDBACKS_STR_NORM_CLIP_VALUE.get(str(self._feedbacks_gen_strategy), torch.inf)
        ))
        self.raise_non_finite_errors = kwargs.get("raise_non_finite_errors", False)

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        """
        Load the state of the optimizer from the checkpoint.

        :param trainer: The trainer object that is used for training.
        :param checkpoint: The checkpoint dictionary.
        :param kwargs: Additional keyword arguments.

        :return: None
        """
        if self.load_state:
            state = checkpoint.get(self.name, {})
            opt_state_dict = state.get(self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY, None)
            if opt_state_dict is not None:
                saved_param_groups = opt_state_dict["param_groups"]
                if self.optimizer is None:
                    self.optimizer = self.DEFAULT_OPTIMIZER_CLS(saved_param_groups)
                self.optimizer.load_state_dict(opt_state_dict)
                self.param_groups = self.optimizer.param_groups
                self.params = self.param_groups[self.OPTIMIZER_PARAMS_GROUP_IDX]["params"]
                self.output_params = self.param_groups[self.OPTIMIZER_OUTPUT_PARAMS_GROUP_IDX]["params"]
                self.feedback_weights = state.get(self.CHECKPOINT_FEEDBACK_WEIGHTS_KEY, None)

    def get_checkpoint_state(self, trainer, **kwargs) -> object:
        """
        Get the state of the optimizer to be saved in the checkpoint.

        :param trainer: The trainer object that is used for training.
        :param kwargs: Additional keyword arguments.
        :return: The state of the optimizer to be saved in the checkpoint.
        """
        if self.save_state:
            state = {}
            if self.optimizer is not None:
                state[self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY] = self.optimizer.state_dict()
                state[self.CHECKPOINT_FEEDBACK_WEIGHTS_KEY] = self.feedback_weights
            return state
        return None

    def _get_feedback_gen_func(self):
        """
        Get the feedbacks generation function.

        :return:
        """
        if self._feedbacks_gen_strategy is None:
            self._feedbacks_gen_strategy = self.DEFAULT_FEEDBACKS_GEN_STRATEGY
        if self._feedbacks_gen_strategy not in self.FEEDBACKS_GEN_FUNCS:
            raise NotImplementedError(
                f"Feedbacks generation strategy '{self._feedbacks_gen_strategy}' is not implemented."
                f"Maybe you meant one of the following: {', '.join(self.FEEDBACKS_GEN_FUNCS.keys())}"
                f" or you can implement your own by adding it to the Eprop.FEEDBACKS_GEN_FUNCS dictionary."
                f" If your new strategy is a common one, please consider contributing to the library by opening a "
                f"pull request on https://github.com/NeuroTorch/NeuroTorch."
            )
        return self.FEEDBACKS_GEN_FUNCS[self._feedbacks_gen_strategy]

    def make_feedback_weights(self, *args, **kwargs):
        """
        Generate the feedback weights for each params.
        The random feedback is noted B_{ij} in Bellec's paper :cite:t:`bellec_solution_2020`.

        :return: The feedback weights for each params.
        """
        feedback_gen_func = self._get_feedback_gen_func()
        feedback_weights = feedback_gen_func(*args, **kwargs)
        clip_tensors_norm_(feedback_weights, max_norm=self.feedback_weights_norm_clip_value)
        return feedback_weights

    def initialize_feedback_weights(
            self,
            y_batch: Optional[Union[Dict[str, torch.Tensor], torch.Tensor]] = None
    ) -> Dict[str, List[torch.Tensor]]:
        """
        TODO : Non-random feedbacks must be implemented with {W_out}.T
        Initialize the feedback weights for each params.
        The random feedback is noted B_{ij} in Bellec's paper :cite:t:`bellec_solution_2020`.
        The keys of the feedback_weights dictionary are the names of the output layers.

        :param y_batch: The batch of the target values.

        :return: The feedback weights for each params.
        """
        if y_batch is None:
            y_batch = self.trainer.current_training_state.y_batch
        if not isinstance(y_batch, dict):
            y_batch = {self.DEFAULT_Y_KEY: y_batch}
        last_dims = [p.shape[-1] if p.ndim > 0 else 1 for p in self.params]
        self.feedback_weights = {
            k: [
                self.make_feedback_weights(y_batch_item.shape[-1], pld, generator=self.rn_gen)
                for pld in last_dims
            ] for k, y_batch_item in y_batch.items()
        }
        return self.feedback_weights

    def initialize_params(self, trainer=None):
        """
        Initialize the parameters of the optimizer.

        :Note: Must be called after :meth:`initialize_output_params` and :meth:`initialize_layers`.

        :param trainer: The trainer to use.

        :return: None
        """
        if not self.params and self.optimizer:
            self.params = self.optimizer.param_groups[self.OPTIMIZER_PARAMS_GROUP_IDX]["params"]

        if not self.params:
            self.params = [
                param
                for layer in self.layers
                for param in layer.parameters()
            ]
        if not self.params:
            warnings.warn("No hidden parameters found. Please provide them manually if you have any.")

        return self.params

    def initialize_output_params(self, trainer=None):
        """
        Initialize the output parameters of the optimizer. Try multiple ways to identify the
        output parameters if those are not provided by the user.

        :Note: Must be called after :meth:`initialize_output_layers`.

        :param trainer: The trainer object.

        :return: None
        """
        if not self.output_params and self.optimizer:
            self.output_params = self.optimizer.param_groups[self.OPTIMIZER_OUTPUT_PARAMS_GROUP_IDX]["params"]

        if not self.output_params:
            self.output_params = [
                param
                for layer in self.output_layers
                for param in layer.parameters()
            ]
        if not self.output_params:
            raise ValueError("Could not find output parameters. Please provide them manually.")

        return

    def initialize_output_layers(self, trainer):
        """
        Initialize the output layers of the optimizer. Try multiple ways to identify the output layers if those are not
        provided by the user.

        :Note: Must be called before :meth:`initialize_output_params`.

        :param trainer: The trainer object.
        :return: None.
        """
        if not self.output_layers:
            self.output_layers = []
            possible_attrs = ["output_layers", "output_layer"]
            for attr in possible_attrs:
                obj = getattr(trainer.model, attr, [])
                if isinstance(obj, (Sequence, torch.nn.ModuleList)):
                    obj = list(obj)
                elif isinstance(obj, (Mapping, torch.nn.ModuleDict)):
                    obj = list(obj.values())
                elif isinstance(obj, torch.nn.Module):
                    obj = [obj]
                self.output_layers += list(obj)

        if not self.output_layers:
            raise ValueError(
                "Could not find output layers. Please provide them manually."
            )

    def initialize_layers(self, trainer):
        """
        Initialize the layers of the optimizer. Try multiple ways to identify the output layers if those are not
        provided by the user.

        :param trainer: The trainer object.

        :return: None
        """
        if not self.layers:
            self.layers = []
            possible_attrs = ["input_layers", "input_layer", "hidden_layers", "hidden_layer"]
            for attr in possible_attrs:
                if hasattr(trainer.model, attr):
                    obj = getattr(trainer.model, attr, [])
                    if isinstance(obj, (Sequence, torch.nn.ModuleList)):
                        obj = list(obj)
                    elif isinstance(obj, (Mapping, torch.nn.ModuleDict)):
                        obj = list(obj.values())
                    elif isinstance(obj, torch.nn.Module):
                        obj = [obj]
                    self.layers += list(obj)
        if not self.layers:
            warnings.warn(
                "No hidden layers found. Please provide them manually if you have any."
                "If you are using only one layer, please note that E-prop is equivalent to a TBPTT. If this is the"
                "case, one might consider using TBPTT instead of E-prop."
            )

    def initialize_param_groups(self) -> List[Dict[str, Any]]:
        """
        The learning rate are initialize. If the user has provided a learning rate for each parameter, then it is used.

        :return: the param_groups.
        """
        self.param_groups = []
        list_insert_replace_at(
            self.param_groups,
            self.OPTIMIZER_PARAMS_GROUP_IDX,
            {"params": self.params, "lr": self._default_params_lr}
        )
        list_insert_replace_at(
            self.param_groups,
            self.OPTIMIZER_OUTPUT_PARAMS_GROUP_IDX,
            {"params": self.output_params, "lr": self._default_output_params_lr}
        )
        return self.param_groups

    def eligibility_traces_zeros_(self):
        """
        Set the eligibility traces to zero.

        :return: None
        """
        self.eligibility_traces = [torch.zeros_like(p) for p in self.params]
        self.output_eligibility_traces = [torch.zeros_like(p) for p in self.output_params]

    def start(self, trainer, **kwargs):
        """
        Start the training process with E-prop.

        :param trainer: The trainer object for the training process with E-prop.
        :param kwargs: Additional arguments.

        :return: None
        """
        LearningAlgorithm.start(self, trainer, **kwargs)
        self.initialize_output_layers(trainer)
        self.initialize_output_params(trainer)
        self.initialize_layers(trainer)
        self.initialize_params(trainer)
        zero_grad_params(self.params)
        zero_grad_params(self.output_params)

        if self.criterion is None and trainer.criterion is not None:
            self.criterion = trainer.criterion

        if not self.param_groups:
            self.initialize_param_groups()
        if not self.optimizer:
            self.optimizer = self.create_default_optimizer()

        self._initialize_original_forwards()
        self.eligibility_traces_zeros_()

    def on_batch_begin(self, trainer, **kwargs):
        """
        For each batch. Initialize the random feedback weights if not already done. Also, set the eligibility traces
        to zero.

        :param trainer: The trainer object.
        :param kwargs: Additional arguments.

        :return: None
        """
        super().on_batch_begin(trainer)
        if trainer.model.training:
            if self.feedback_weights is None:
                self.initialize_feedback_weights(trainer.current_training_state.y_batch)
        self.eligibility_traces_zeros_()
        zero_grad_params(self.params)
        zero_grad_params(self.output_params)

    def decorate_forwards(self):
        """
        Ensure that the forward pass is decorated. THe original forward and the hidden layers names are stored. The
        hidden layers forward method are decorated using :meth: `_decorate_hidden_forward`. The output layers forward
        are decorated using :meth: `_decorate_output_forward` from TBPTT.

        Here, we are using decorators to introduce a specific behavior in the forward pass. For E-prop, we need to
        ensure that the gradient is computed and optimize at each time step t of the sequence. This can be achieved by
        decorating our forward. However, we do keep in storage the previous forward pass. This is done to ensure
        that the forward pass is not modified permanently in any way.


        :return: None
        """
        if self.trainer.model.training:
            if not self._forwards_decorated:
                self._initialize_original_forwards()
            self._hidden_layer_names.clear()

            for layer in self.layers:
                self._hidden_layer_names.append(layer.name)
                if self._use_hooks:
                    hook = layer.register_forward_hook(self._hidden_hook, with_kwargs=True)
                    self.forwards_hooks.append(hook)
                else:
                    layer.forward = self._decorate_hidden_forward(layer.forward, layer.name)

            for layer in self.output_layers:
                if self._use_hooks:
                    hook = layer.register_forward_hook(self._output_hook, with_kwargs=True)
                    self.forwards_hooks.append(hook)
                else:
                    layer.forward = self._decorate_forward(layer.forward, layer.name)
            self._forwards_decorated = True

    def _decorate_hidden_forward(self, forward, layer_name: str) -> Callable:
        """
        In TBPTT, we decorate forward to ensure that the backpropagation and the optimizer at t is done for the entire
        network. In E-prop, we want to backpropagate the hidden layers locally (and not the entire network) at each
        time step t.

        :param forward: The forward method to decorate.
        :param layer_name: The name of the layer.

        :return: The decorated forward method
        """
        def _forward(*args, **kwargs):
            out = forward(*args, **kwargs)
            t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
            if t is None:
                return out
            out_tensor, hh = unpack_out_hh(out)
            list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
            if len(self._layers_buffer[layer_name]) >= self.backward_time_steps:
                self._hidden_backward_at_t(t, self.backward_time_steps, layer_name)
                out = recursive_detach(out)
            return out
        return _forward

    def _hidden_hook(self, module, args, kwargs, output) -> None:
        r"""
        In TBPTT, we decorate forward to ensure that the backpropagation and the optimizer at t is done for the entire
        network. In E-prop, we want to backpropagate the hidden layers locally (and not the entire network) at each
        time step t.
        """
        t, forecasting = kwargs.get("t", None), kwargs.get("forecasting", False)
        if t is None:
            return

        layer_name = module.name
        out_tensor, hh = unpack_out_hh(output)
        list_insert_replace_at(self._layers_buffer[layer_name], t % self.backward_time_steps, out_tensor)
        if len(self._layers_buffer[layer_name]) >= self.backward_time_steps:
            self._hidden_backward_at_t(t, self.backward_time_steps, layer_name)
            output = recursive_detach_(output)
        return

    def _hidden_backward_at_t(self, t: int, backward_t: int, layer_name: str):
        """
        Here, we compute the eligibility trace as seen in equation (13) from :cite:t:`bellec_solution_2020`. Please
        note that while the notation used in this paper for the equation (13) is [dz/dW]_{local}, we have used [dy/dW]
        in order to be coherent with our own convention.

        :param t: The current time step. For example, if we have a time series of length 100, t will be in [0, 99].
        :param backward_t: The number of time steps to go back in time.
        :param layer_name: The name of the layer.

        :return: None
        """
        pred_batch = torch.squeeze(self._get_pred_batch_from_buffer(layer_name))
        dy_dw_locals = dy_dw_local(y=pred_batch, params=self.params, retain_graph=True, allow_unused=True)
        with torch.no_grad():
            self.eligibility_traces = [
                self.gamma * et + torch.nan_to_num(
                    dy_dw.to(et.device),
                    nan=0.0,
                    neginf=-self.eligibility_traces_norm_clip_value,
                    posinf=self.eligibility_traces_norm_clip_value,
                )
                for et, dy_dw in zip(self.eligibility_traces, dy_dw_locals)
            ]
            clip_tensors_norm_(self.eligibility_traces, self.eligibility_traces_norm_clip_value)
            self._layers_buffer[layer_name].clear()

    def _backward_at_t(self, t: int, backward_t: int, layer_name: str):
        """
        Apply the criterion on the batch. The gradients of each parameters are then updated but are not yet optimized.

        :param t: current time step
        :param backward_t: number of time steps to go back in time
        :param layer_name: name of the layer

        :return: None
        """
        y_batch = self._get_y_batch_slice_from_trainer((t + 1) - backward_t, t + 1, layer_name)
        pred_batch = self._get_pred_batch_from_buffer(layer_name)
        batch_loss = self.apply_criterion(pred_batch, y_batch)
        if batch_loss.grad_fn is None:
            raise ValueError(
                f"batch_loss.grad_fn is None. This is probably an internal error. Please report this issue on GitHub."
            )
        with torch.no_grad():
            errors = self.compute_errors(pred_batch, y_batch)
        output_grads = dy_dw_local(torch.mean(batch_loss), self.output_params, retain_graph=True, allow_unused=True)
        with torch.no_grad():
            self.output_eligibility_traces = [
                self.alpha * et + torch.nan_to_num(
                    dy_dw.to(et.device),
                    nan=0.0,
                    neginf=-self.eligibility_traces_norm_clip_value,
                    posinf=self.eligibility_traces_norm_clip_value,
                )
                for et, dy_dw in zip(self.output_eligibility_traces, output_grads)
            ]
            clip_tensors_norm_(self.output_eligibility_traces, self.eligibility_traces_norm_clip_value)
        self.update_grads(errors)
        with torch.no_grad():
            self._layers_buffer[layer_name].clear()

    def compute_learning_signals(self, errors: Dict[str, torch.Tensor]) -> List[torch.Tensor]:
        """
        TODO : Determine if we normalize with the number of output when computing the learning signal : If multiple
        TODO : output layers, do we sum the learning signals or do we average them ? Should we make a 'reduce' param?
        TODO : When averaging, add factor 1/n to the learning signal. It "kind of" results in a change of learning rate.
        The learning signals are computed using equation (28) from :cite:t:`bellec_solution_2020`.

        :param errors: The errors to use to compute the learning signals.

        :return: List of the learning signals for each parameter.
        """
        learning_signals = [torch.zeros((p.shape[-1] if p.ndim > 0 else 1), device=p.device) for p in self.params]
        for k, feedbacks in self.feedback_weights.items():
            if k not in errors:
                raise ValueError(
                    f"This is an internal error. Please report this issue on GitHub."
                    f"Key {k} from {self.feedback_weights.keys()=} not found in errors of keys {errors.keys()}."
                )
            torch.nan_to_num_(errors[k], nan=self.nan, posinf=self.posinf, neginf=self.neginf)
            error_mean = torch.mean(errors[k].view(-1, errors[k].shape[-1]), dim=0).view(1, -1)
            if torch.isfinite(self.learning_signal_norm_clip_value):
                clip_tensors_norm_(error_mean, max_norm=self.learning_signal_norm_clip_value)
            for i, feedback in enumerate(feedbacks):
                learning_signals[i] = learning_signals[i] + torch.matmul(error_mean, feedback.to(error_mean.device))
        return learning_signals

    def compute_errors(
            self,
            pred_batch: Union[Dict[str, torch.Tensor], torch.Tensor],
            y_batch: Union[Dict[str, torch.Tensor], torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """
        The errors for each output is computed then inserted in a dict for further use. This function check if the
        y_batch and pred_batch are given as a dict or a tensor.

        :param pred_batch: prediction of the network
        :param y_batch: target

        :return: dict of errors
        """
        if isinstance(y_batch, dict) or isinstance(pred_batch, dict):
            if isinstance(y_batch, torch.Tensor):
                y_batch = {k: y_batch for k in pred_batch}
            else:
                raise ValueError(f"y_batch must be a dict or a tensor, not {type(y_batch)}.")
            if isinstance(pred_batch, torch.Tensor):
                pred_batch = {k: pred_batch for k in y_batch}
            else:
                raise ValueError(f"pred_batch must be a dict or a tensor, not {type(pred_batch)}.")
            batch_errors = {
                k: (pred_batch[k] - y_batch[k].to(pred_batch[k].device))
                for k in y_batch
            }
        else:
            batch_errors = {self.DEFAULT_Y_KEY: pred_batch - y_batch.to(pred_batch.device)}
        return batch_errors

    def update_grads(
            self,
            errors: Dict[str, torch.Tensor],
    ):
        """
        The learning signal is computed. The gradients of the parameters are then updated as seen in equation (28)
        from :cite:t:`bellec_solution_2020`.

        :param errors: The errors to use to compute the learning signals.

        :return: None
        """
        learning_signals = self.compute_learning_signals(errors)
        with torch.no_grad():
            for param, ls, et in zip(self.params, learning_signals, self.eligibility_traces):
                if param.requires_grad:
                    param.grad += (ls * et.to(ls.device)).to(param.device).view(param.shape).detach()
                    torch.nan_to_num_(param.grad, nan=self.nan, posinf=self.posinf, neginf=self.neginf)
            if torch.isfinite(self.grad_norm_clip_value):
                torch.nn.utils.clip_grad_norm_(self.params, self.grad_norm_clip_value)

            if self.raise_non_finite_errors:
                if not all([torch.isfinite(p.grad).all() for p in self.params]):
                    raise ValueError(
                        "Non-finite detected in hidden parameters gradients. Try to reduce the learning rate of the hidden "
                        "parameters with the argument `params_lr`."
                    )

        with torch.no_grad():
            for out_param, out_el in zip(self.output_params, self.output_eligibility_traces):
                if out_param.requires_grad:
                    out_param.grad += out_el.to(out_param.device).view(out_param.shape).detach()
                    torch.nan_to_num_(out_param.grad, nan=self.nan, posinf=self.posinf, neginf=self.neginf)
            if torch.isfinite(self.grad_norm_clip_value):
                torch.nn.utils.clip_grad_norm_(self.output_params, self.grad_norm_clip_value)

            if self.raise_non_finite_errors:
                if not all([torch.isfinite(p).all() for p in self.output_params]):
                    raise ValueError(
                        "Non-finite detected in output parameters gradients. Try to reduce the learning rate of the output "
                        "parameters with the argument `output_params_lr`."
                    )

    def _make_optim_step(self, **kwargs):
        """
        Set the gradients and the eligibility traces to zero.

        :param kwargs: Additional arguments.

        :return: None
        """
        super()._make_optim_step(**kwargs)
        with torch.no_grad():
            zero_grad_params(self.output_params)
        if self.raise_non_finite_errors:
            if not all([torch.isfinite(p).all() for p in self.params]):
                raise ValueError(
                    "Non-finite detected in hidden parameters. Try to reduce the learning rate of the hidden "
                    "parameters with the argument `params_lr`."
                )
            if not all([torch.isfinite(p).all() for p in self.output_params]):
                raise ValueError(
                    "Non-finite detected in output parameters. Try to reduce the learning rate of the output "
                    "parameters with the argument `output_params_lr`."
                )

    def on_batch_end(self, trainer, **kwargs):
        """
        Ensure that there is not any remaining gradients in the output parameters. The forward are undecorated and the
        gradients are set to zero. The buffer are also cleared.

        :param trainer: The trainer to use for computation.
        :param kwargs: Additional arguments.

        :return: None
        """
        LearningAlgorithm.on_batch_end(self, trainer)
        if trainer.model.training:
            need_optim_step = False
            for layer_name in self._layers_buffer:
                backward_t = len(self._layers_buffer[layer_name])
                if backward_t > 0:
                    need_optim_step = True
                    if layer_name in self._hidden_layer_names:
                        self._hidden_backward_at_t(self._data_n_time_steps - 1, backward_t, layer_name)
                    else:
                        self._backward_at_t(self._data_n_time_steps - 1, backward_t, layer_name)
            if need_optim_step:
                self._make_optim_step()
        self.undecorate_forwards()
        self._layers_buffer.clear()
        self.optimizer.zero_grad()
        self.eligibility_traces_zeros_()





import os
from copy import deepcopy
from typing import Optional, Tuple, Any, Sequence

import matplotlib.pyplot as plt

import numpy as np
import torch

from matplotlib import animation
from scipy import interpolate
from sklearn.cluster import KMeans, DBSCAN
from sklearn.decomposition import PCA

from ..dimension import DimensionProperty, Size, DimensionsLike
from ..metrics import PVarianceLoss
from ..transforms.base import to_numpy, to_tensor


class Visualise:
    """
    This class is used to visualise the time series without any form of clustering.
    The time series can be visualised in the following ways:
        1. Animate the time series
        2. Heatmap of the time series
        3. Rigid plot of the time series
        4. Plot all the neuronal activity in one figure
    Further visualisation are already added. You can visualise the time series with clustering methods. You can
    then visualise those clustered time series in scatter/trajectory in their clustered space.
    """

    @staticmethod
    def number_axes(axes: Sequence[plt.Axes], **kwargs) -> Sequence[plt.Axes]:
        """
        Number the axes.

        :param axes: Axes to number.
        :type axes: Sequence[plt.Axes]
        :param kwargs: Keyword arguments.

        :keyword str num_type: Type of number to display. Can be either "alpha" or "numeric".
        :keyword int start: Number to start with.
        :keyword float x: x position of the number in the axes coordinate (see ax.transAxes). Default is 0.0.
        :keyword float y: y position of the number in the axes coordinate (see ax.transAxes). Default is 1.2.
        :keyword float fontsize: Font size of the number. Default is 12.
        :keyword str fontweight: Font weight of the number. Default is "bold".
        :keyword method: Method to use to number the axes. Available methods are "text", "title" and "set_title".
            The "text" method will add a text to the axes. The "title" method will add the number to the existing title.
            The "set_title" method will set the title of the axes, so the existing title will be overwritten.
            Default is "text".

        :return: The axes with the number.
        :rtype: Sequence[plt.Axes]
        """
        axes_view = np.ravel(np.asarray(axes))
        num_type = kwargs.get("num_type", "alpha").lower()
        mth = kwargs.get("method", "text").lower()
        start = kwargs.get("start", 0)
        if num_type == "alpha":
            axes_numbers = [chr(i) for i in range(97 + start, 97 + len(axes_view) + start)]
        elif num_type == "numeric":
            axes_numbers = [str(i) for i in range(1 + start, len(axes_view) + 1 + start)]
        else:
            raise ValueError(f"Unknown num_type {num_type}.")
        for i, ax in enumerate(axes_view):
            if mth == "text":
                ax.text(
                    kwargs.get("x", 0.0), kwargs.get("y", 1.2),
                    f"({axes_numbers[i]})",
                    transform=ax.transAxes, fontsize=kwargs.get("fontsize", 12),
                    fontweight=kwargs.get("fontweight", 'bold'), va='top'
                )
            elif mth == "title":
                ax.set_title(
                    f"({axes_numbers[i]}) {ax.get_title()}",
                    fontsize=kwargs.get("fontsize", 12),
                    fontweight=kwargs.get("fontweight", 'bold'),
                    loc=kwargs.get("loc", "left"),
                )
            elif mth == "set_title":
                ax.set_title("")
                ax.set_title(
                    f"({axes_numbers[i]})",
                    fontsize=kwargs.get("fontsize", 12),
                    fontweight=kwargs.get("fontweight", 'bold'),
                    loc=kwargs.get("loc", "left"),
                )
        return axes

    def __init__(
            self,
            timeseries: Any,
            shape: Optional[DimensionsLike] = None,
            apply_zscore: bool = False
    ):
        """
        :param timeseries: Time series of shape (Batch size, Time Steps, Features).
        :param shape: Shape of the time series. If shape is None, the shape is inferred from the time series. Useful
        to identify the dtype of the time series. If the shape is Size, make sure to set the name of the dimensions as
        you want them to be displayed.

        :Note: If the times series is 3+ dimensional, the first dimension is assumed to be the batch size.
            The time series will be mean over the batch size.
        """
        self.timeseries = to_numpy(timeseries)
        self._given_timeseries = deepcopy(self.timeseries)
        self.given_shape_mean = (-1, *self.timeseries.shape[-2:])
        if len(self.timeseries.shape) >= 3:
            self.timeseries = np.mean(self.timeseries.reshape(self.given_shape_mean), axis=0)
            self._mean_given_timeseries = np.mean(self._given_timeseries.reshape(self.given_shape_mean), axis=0)
            self._std_given_timeseries = np.std(self._given_timeseries.reshape(self.given_shape_mean), axis=0)
            self.is_mean = True
        else:
            self._mean_given_timeseries = self._given_timeseries
            self._std_given_timeseries = np.zeros_like(self._given_timeseries)
            self.is_mean = False

        self._is_transposed = False
        self.shape = self._set_dimension(shape)
        self._transpose_given_time_series_if_needed()
        if apply_zscore:
            self._zscore_timeseries()

    def _transpose_given_time_series_if_needed(self):
        if self._is_transposed:
            if self.is_mean:
                self._given_timeseries = self._given_timeseries.transpose(0, 2, 1)
            else:
                self._given_timeseries = self._given_timeseries.transpose(1, 0)

    def _zscore_timeseries(self):
        """
        Z-score the time series.
        """
        for i in range(int(self.shape[-1])):
            self.timeseries[:, i] = (
                    (self.timeseries[:, i] - np.mean(self.timeseries[:, i])) / np.std(self.timeseries[:, i])
            )

    def _set_dimension(self, shape: Optional[DimensionsLike]) -> Size:
        """
        Identify the shape of the time series. Will transpose the time series to have the time dimension as the first
        dimension.

        :param shape: Shape of the time series. Use object Size given in this package.
        :return: The shape of the time series.
        """
        if shape is None:
            shape = self.timeseries.shape

        shape = Size(shape)
        assert len(shape) == 2, "The shape of the time series must be 2 dimensional. If the time series includes a " \
                                "batch dimension, the batch must not be included in the shape. "
        if all(dim.dtype == DimensionProperty.NONE for dim in shape):
            shape[0].dtype = DimensionProperty.TIME
            for dim in shape:
                if dim.dtype == DimensionProperty.TIME:
                    dim.name = "Time Steps"
                elif dim.dtype == DimensionProperty.NONE:
                    dim.name = "Features"
        for i, dim in enumerate(shape):
            if dim.size is None:
                dim.size = self.timeseries.shape[i]

        if shape[0].dtype == DimensionProperty.NONE:
            self.timeseries = self.timeseries.T
            self._is_transposed = True
            shape = Size(shape.dimensions[::-1])
        return shape

    def animate(
            self,
            weights: np.ndarray,
            dt: float = 1.0,
            step: int = 1,
            time_interval: float = 1.0,
            node_size: float = 50,
            alpha: Optional[float] = None,
            filename: Optional[str] = None,
            file_extension: Optional[str] = None,
            show: bool = False,
            **kwargs
    ):
        """
        Animate the time series. The position of the nodes are obtained using the spring layout.
        Spring-Layout use the Fruchterman-Reingold force-directed algorithm. For more information,
        please refer to the following documentation of networkx:
        https://networkx.org/documentation/stable/reference/generated/networkx.drawing.layout.spring_layout.html

        :param weights: Weight matrix of size (number of neurons, number of neurons).
        :param dt: Time step between two time steps.
        :param step: Number of time step between two animation frames.
            example: if step = 4, the animation will play at t = 0, t = 4, t = 8, t = 12 ...
        :param time_interval: Time interval between two animation frames (in milliseconds)
        :param node_size: Size of the nodes
        :param alpha: Density of the connections. Small network should have a higher alpha value.
        :param filename: Name of the file to save the animation. If filename is None, the animation will not be saved.
        The application imagemagick is required to save the animation.
        :param file_extension: Extension of the file to save the animation. The available extensions are: mp4 and gif.
        :param show: If True, the animation will be displayed.
        :param kwargs: Keyword arguments.

        :keyword fps: Frames per second. Default is 30.
        """
        import networkx as nx

        if alpha is None:
            alpha = 1.0 / np.sqrt(weights.size)
        num_frames = int(self.shape[0]) // step
        connectome = nx.from_numpy_array(weights)
        pos = nx.spring_layout(connectome)
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (14, 8)))
        nx.draw_networkx_nodes(
            connectome,
            pos,
            ax=ax,
            node_size=node_size,
            node_color=self.timeseries[0],
            cmap="hot"
        )
        nx.draw_networkx_edges(connectome, pos, ax=ax, width=1.0, alpha=alpha)
        plt.axis("off")
        text = ax.text(0.5, 0.97, rf"$t = 0 / {int(self.shape[0]) * dt}$", ha="center", transform=ax.transAxes)
        plt.tight_layout(pad=0)

        def _animation(i):
            nodes = nx.draw_networkx_nodes(
                connectome,
                pos,
                ax=ax, node_size=node_size,
                node_color=self.timeseries[i * step],
                cmap="hot"
            )
            text.set_text(rf"$t = {i * step * dt:.3f} / {int(self.shape[0]) * dt}$")
            return nodes, text

        anim = animation.FuncAnimation(fig, _animation, frames=num_frames, interval=time_interval, blit=True)
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            if file_extension is None:
                if '.' in filename:
                    file_extension = filename.split('.')[-1]
                else:
                    file_extension = 'gif'
            assert file_extension in ["mp4", "gif"], "The extension of the file must be mp4 or gif."
            if filename.endswith(file_extension):
                filename = ''.join(filename.split('.')[:-1])
            anim.save(
                f"{filename}.{file_extension}",
                writer=kwargs.get("writer", "imagemagick"),
                fps=kwargs.get("fps", 30)
            )
        if show:
            plt.show()

    def plot_timeseries(
            self,
            filename: Optional[str] = None,
            show: bool = False,
            fig: Optional[plt.Figure] = None,
            ax: Optional[plt.Axes] = None,
            **kwargs
    ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot all the neuronal activity in one figure.

        :param filename: Name of the file to save the figure. If filename is None, the figure will not be saved.
        :param show: If True, the figure will be displayed.
        :param fig: Figure to plot the timeseries. If fig is None, a new figure will be created.
        :param ax: Axes to plot the timeseries. If ax is None, a new axes will be created.
        :param kwargs: Keyword arguments.

        :keyword figsize: Size of the figure. Default is (12, 8).
        :keyword dpi: DPI of the figure. Default is 300.
        """
        assert (fig is None) == (ax is None), "fig and ax must be both None or both not None."
        if fig is None or ax is None:
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", (12, 8)))
        ax.set_xlabel(self.shape[0].name)
        ax.plot(self.timeseries)
        ax.set_ylabel(self.shape[1].name)
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            fig.savefig(filename, dpi=kwargs.get("dpi", 300))
        if show:
            plt.show()
        return fig, ax

    def heatmap(
            self,
            show_axis: bool = True,
            interpolation: str = "nearest",
            cmap: str = "RdBu_r",
            v: Tuple[float, float] = (0.0, 1.0),
            filename: Optional[str] = None,
            show: bool = False,
            fig: Optional[plt.Figure] = None,
            ax: Optional[plt.Axes] = None,
            **kwargs
    ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot the heatmap of the time series.

        :param show_axis: Whether to show the axis or not.
        :param interpolation: Type of interpolation between the time step.
        :param cmap: Colormap of the heatmap.
        :param v: Range of the colorbar.
        :param filename: Name of the file to save the figure. If filename is None, the figure will not be saved.
        :param show: If True, the figure will be displayed.
        :param fig: Figure to plot the heatmap. If fig is None, a new figure will be created.
        :param ax: Axes to plot the heatmap. If ax is None, a new axes will be created.
        :param kwargs: Keyword arguments.

        :keyword figsize: Size of the figure. Default is (12, 8).
        :keyword dpi: DPI of the figure. Default is 300.
        """
        assert (fig is None) == (ax is None), "fig and ax must be both None or both not None."
        if fig is None or ax is None:
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", (12, 8)))
        ax.set_xlabel(self.shape[0].name)
        ax.set_ylabel(self.shape[1].name)
        im = ax.imshow(self.timeseries.T, interpolation=interpolation, aspect="auto", cmap=cmap, vmin=v[0], vmax=v[1])
        if not show_axis:
            ax.axis("off")
        ax.set_title(kwargs.get("title", ""))
        cbar = fig.colorbar(im)
        cbar.set_label('Activity [-]', rotation=270)
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            fig.savefig(filename, dpi=kwargs.get("dpi", 300))
        if show:
            plt.show()
        return fig, ax

    def rigidplot(
            self,
            show_axis: bool = False,
            filename: Optional[str] = None,
            show: bool = False,
            fig: Optional[plt.Figure] = None,
            ax: Optional[plt.Axes] = None,
            **kwargs
    ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot the rigid plot of the time series.

        :param show_axis: Whether to show the axis or not.
        :param filename: Name of the file to save the figure. If filename is None, the figure will not be saved.
        :param show: If True, the figure will be displayed.
        :param fig: Figure to plot the rigid plot. If fig is None, a new figure will be created.
        :param ax: Axes to plot the rigid plot. If ax is None, a new axes will be created.
        :param kwargs: Keyword arguments.

        :keyword figsize: Size of the figure. Default is (12, 8).
        :keyword dpi: DPI of the figure. Default is 300.
        """
        assert (fig is None) == (ax is None), "fig and ax must be both None or both not None."
        if fig is None or ax is None:
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", (12, 8)))
        ax.set_xlabel(self.shape[0].name)
        ax.set_ylabel(self.shape[1].name)
        if not show_axis:
            ax.axis("off")
        for i in range(int(self.shape[-1])):
            shifted_timeseries = self.timeseries[:, i] - np.min(self.timeseries[:, i])
            shifted_timeseries = np.divide(
                shifted_timeseries, np.max(shifted_timeseries),
                out=np.zeros_like(shifted_timeseries),
                where=np.logical_not(np.isclose(np.max(shifted_timeseries), 0))
            ) + 0.8 * i
            ax.plot(shifted_timeseries, c="k", alpha=0.9, linewidth=1.0)
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            fig.savefig(filename, dpi=kwargs.get("dpi", 300))
        if show:
            plt.show()
        return fig, ax

    def plot_single_timeseries_comparison(
            self,
            feature_index: int,
            ax: plt.Axes,
            target: Any,
            spikes: Optional[Any] = None,
            n_spikes_steps: Optional[int] = None,
            title: str = "",
            desc: str = "Predicted time series",
            **kwargs
    ) -> plt.Axes:
        predictions, target = to_tensor(self._mean_given_timeseries[:, feature_index]), to_tensor(target)
        if self.is_mean:
            ax.fill_between(
                np.arange(predictions.shape[0]),
                to_numpy(predictions) - to_numpy(self._std_given_timeseries[:, feature_index]),
                to_numpy(predictions) + to_numpy(self._std_given_timeseries[:, feature_index]),
                alpha=0.2, color="blue"
            )

        pVar = PVarianceLoss()(predictions, target)
        ax.plot(to_numpy(predictions), label=f"{desc} (pVar: {to_numpy(pVar).item():.3f})", c="tab:blue",
                linewidth=kwargs.get("linewidth", 2.5))
        ax.plot(to_numpy(target), label="Real time series", c="tab:orange", linewidth=kwargs.get("linewidth", 2.5))

        if "box_aspect" in kwargs:
            ax.set_box_aspect(kwargs.get("box_aspect"))

        if spikes is not None:
            spikes = to_tensor(spikes)
            assert len(spikes.shape) == 2, "spikes must be a 2D tensor"
            assert n_spikes_steps is not None, "n_spikes_steps must be provided if spikes is not None"
            y_max = max(target.max(), predictions.max())
            x_scatter_space = np.linspace(0, len(target), num=n_spikes_steps * len(target))
            x_scatter_spikes = []
            x_scatter_values = []
            x_scatter_zeros = []
            for i, xs in enumerate(x_scatter_space):
                if np.isclose(
                        spikes[i // n_spikes_steps][i % n_spikes_steps],
                        0.0
                ):
                    x_scatter_zeros.append(xs)
                else:
                    x_scatter_spikes.append(xs)
                    x_scatter_values.append(spikes[i // n_spikes_steps][i % n_spikes_steps])
            x_scatter_values = np.clip(x_scatter_values, 0.0, 1.0)
            if x_scatter_values.size == 0:
                x_scatter_values = 0.0
            ax.scatter(
                x_scatter_spikes, y=[y_max * 1.1] * len(x_scatter_spikes),
                label="Latent space", c='k', marker='|', linewidths=0.5, alpha=x_scatter_values,
            )

        ax.set_xlabel(self.shape[0].name)
        ax.set_ylabel(kwargs.get("ylabel", self.shape[1].name))
        ax.set_title(title)
        if spikes is not None:
            ax.legend(loc=kwargs.get("legend_loc", "upper right"),
                      bbox_to_anchor=kwargs.get("legend_bbox_to_anchor", (1, 0.93)))
        else:
            ax.legend(loc=kwargs.get("legend_loc", "upper right"))
        return ax

    def plot_timeseries_comparison(
            self,
            target: Any,
            spikes: Optional[Any] = None,
            n_spikes_steps: Optional[int] = None,
            title: str = "",
            desc: str = "Prediction",
            filename: Optional[str] = None,
            show: bool = False,
            fig: Optional[plt.Figure] = None,
            axes: Optional[Sequence[plt.Axes]] = None,
            **kwargs
    ) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
        """
        Plot the timeseries comparison.

        :param target: Target timeseries. Must be in shape (n_time_steps, n_features).
        :param spikes: Spikes of the latent space. Must be in shape (n_time_steps, n_spikes_steps, n_spikes).
        :param n_spikes_steps: Number of spikes steps. If None, spikes must be None.
        :param title: Title of the plot.
        :param desc: Description of the prediction.
        :param filename: Filename to save the plot to.
        :param show: Whether to show the plot.
        :param fig: Figure to plot on. If None, a new figure is created.
        :param axes: Axes to plot on. If None, new axes are created.
        :param kwargs: Additional keyword arguments.

        :keyword bool close: Whether to close the figure after saving. Default: False.
        :keyword int dpi: DPI of the figure. Default: 300.
        :keyword str legend_loc: Location of the legend. Default: "upper right".
        :keyword List[str] traces_to_show: List of traces to show.
                Available: ["error_quad", "best", "most_var", "worst", "typical_<int>"]
                where <int> is the index of the typical trace,
                starting from 0. The typical ones refers to the ones that are closest to the mean pVar.
                Default: ["error_quad", "best", "most_var", "worst"].
        :keyword bool numbered: Whether to number the traces. Default: False.

        :return: Figure and axes.
        """
        assert (fig is None) == (axes is None), "fig and axes must be both None or both not None"
        predictions, target = to_tensor(self._mean_given_timeseries), to_tensor(target)
        target = torch.squeeze(target.detach().cpu())

        errors = torch.squeeze(predictions - target.to(predictions.device)) ** 2
        if self.is_mean:
            mean_pVar, std_pVar = PVarianceLoss().mean_std_over_batch(self._given_timeseries, target[np.newaxis, ...])
            mean_pVar, std_pVar = to_numpy(mean_pVar).item(), to_numpy(std_pVar).item()
            batch_size = self._given_timeseries.shape[0]
            title = f"{title} (pVar[{batch_size}]: {mean_pVar:.3f} ± {std_pVar:.3f})"
        else:
            pVar = PVarianceLoss()(predictions, target.to(predictions.device))
            title = f"{title} (pVar: {to_numpy(pVar).item():.3f})"

        pVar_per_feature = PVarianceLoss(reduction='feature')(predictions, target.to(predictions.device))
        mean_pVar = PVarianceLoss(reduction='mean')(predictions, target.to(predictions.device))
        mean_pVar_sort, indices = torch.sort(pVar_per_feature, descending=True)
        var_diff, var_diff_indices = torch.sort(torch.var(torch.diff(target, dim=0), dim=0), descending=True)
        typical_pVar_sort, typical_indices = torch.sort(torch.abs(pVar_per_feature - mean_pVar), descending=False)
        target = torch.squeeze(target).numpy().T

        traces_to_show = kwargs.get("traces_to_show", ["error_quad", "best", "most_var", "worst"])
        traces_to_show = [t.lower() for t in traces_to_show]
        plot_error_quad = "error_quad" in traces_to_show
        traces_to_indexes = {
            "error_quad": None,
            "best": indices[0],
            "worst": indices[-1],
            "most_var": var_diff_indices[0],
        }
        traces_to_names = {
            "error_quad": "Squared Error [-]",
            "best": "Best",
            "worst": "Worst",
            "most_var": "Most Var",
        }
        for i, trace in enumerate(traces_to_show):
            traces_to_show[i] = trace.lower()
            if "typical" in trace:
                typical_idx = int(trace.split("_")[1])
                traces_to_indexes[trace] = typical_indices[typical_idx]
                traces_to_names[trace] = f"Typical {typical_idx}"
            if trace not in traces_to_indexes:
                raise ValueError(f"Unknown trace to show: {trace}. Known traces: {list(traces_to_indexes.keys())}")

        given_names = kwargs.get("traces_to_show_names", [traces_to_names[t] for t in traces_to_show])
        assert len(given_names) == len(
            traces_to_show), "traces_to_show_names must have the same length as traces_to_show"
        for trace, trace_name in zip(traces_to_show, given_names):
            traces_to_names[trace] = trace_name
        if plot_error_quad:
            traces_to_show.remove("error_quad")

        n_plot = len(traces_to_show) + int(plot_error_quad)
        if fig is None or axes is None:
            fig, axes = plt.subplots(n_plot, 1, figsize=kwargs.get("figsize", (16, 8)))
        else:
            assert len(axes) == n_plot, f"axes must have length {n_plot}"
        if plot_error_quad:
            axes[0].plot(to_numpy(errors))
            axes[0].set_xlabel(self.shape[0].name)
            axes[0].set_ylabel(traces_to_names["error_quad"])
            axes[0].set_title(title)

        if spikes is not None:
            spikes = to_numpy(spikes)
            trace_to_spikes_indexes = {
                trace_name: spikes[:, :, indexes]
                for trace_name, indexes in traces_to_indexes.items()
            }
        else:
            trace_to_spikes_indexes = {}
        for i, trace in enumerate(traces_to_show):
            self.plot_single_timeseries_comparison(
                traces_to_indexes[trace], axes[i + int(plot_error_quad)], target[traces_to_indexes[trace]],
                trace_to_spikes_indexes.get(trace, None),
                n_spikes_steps=n_spikes_steps,
                title=f"{traces_to_names[trace]}", desc=desc,
                **kwargs
            )

        fig.set_tight_layout(True)
        if kwargs.get("numbered", False):
            self.number_axes(axes)
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            fig.savefig(filename, dpi=kwargs.get("dpi", 300))
        if show:
            plt.show()
        if kwargs.get("close", False):
            plt.close(fig)
        return fig, axes

    def plot_timeseries_comparison_report(
            self,
            target: Any,
            spikes: Optional[Any] = None,
            n_spikes_steps: Optional[int] = None,
            title: str = "",
            desc: str = "Prediction",
            filename: Optional[str] = None,
            show: bool = False,
            fig: Optional[plt.Figure] = None,
            axes: Optional[Sequence[plt.Axes]] = None,
            **kwargs
    ) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
        if fig is None or axes is None:
            fig, axes = plt.subplots(ncols=2, nrows=4, figsize=kwargs.get("figsize", (16, 8)))
        else:
            axes = np.asarray(axes)
            assert axes.shape == (4, 2), f"axes must have shape (4, 2), got {axes.shape}"
        gs = axes[0, 0].get_gridspec()
        for ax in axes[0, :]:
            ax.remove()
        axbig = fig.add_subplot(gs[0, :])
        self.plot_timeseries_comparison(
            target, spikes,
            n_spikes_steps=n_spikes_steps,
            title=title, desc=desc,
            fig=fig, axes=[axbig],
            traces_to_show=["error_quad", ],
            traces_to_show_names=["Squared error [-]", ],
            show=False,
        )
        self.plot_timeseries_comparison(
            target, spikes,
            n_spikes_steps=n_spikes_steps,
            title=title, desc=desc,
            fig=fig, axes=axes[1:, 0],
            traces_to_show=["best", "most_var", "worst"],
            traces_to_show_names=[
                "Best Neuron Prediction", "Most variable Neuron Prediction", "Worst Neuron Prediction"
            ],
            show=False,
        )
        self.plot_timeseries_comparison(
            target, spikes,
            n_spikes_steps=n_spikes_steps,
            title=title, desc=desc,
            fig=fig, axes=axes[1:, 1],
            traces_to_show=[f"typical_{i}" for i in range(3)],
            traces_to_show_names=[
                "Typical Neuron Prediction (1)",
                "Typical Neuron Prediction (2)",
                "Typical Neuron Prediction (3)"
            ],
            show=False,
        )
        self.number_axes([axbig] + list(np.ravel(axes[1:])), **kwargs)
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            fig.savefig(filename, dpi=kwargs.get("dpi", 300))
        if show:
            plt.show()
        if kwargs.get("close", False):
            plt.close(fig)
        return fig, axes


class VisualiseKMeans(Visualise):
    """
    Visualise the time series using only a K-means algorithm of clustering.
    """

    def __init__(
            self,
            timeseries: Any,
            shape: Optional[DimensionsLike] = None,
            apply_zscore: bool = False,
            n_clusters: int = 13,
            random_state: int = 0
    ):
        """
        Constructor of the class.

        :param timeseries: Time series of shape (n_time_steps, n_neurons).
        :param shape: Shape of the time series. If shape is None, the shape is inferred from the time series. Useful
        to identify the dtype of the time series. If the shape is Size, make sure to set the name of the dimensions as
        you want them to be displayed.
        :param apply_zscore: Whether to apply z-score or not.
        :param n_clusters: Number of clusters.
        :param random_state: Determines random number generation for centroid initialization.
        Use an int to make the randomness deterministic
        """
        super().__init__(
            timeseries=timeseries,
            shape=shape,
            apply_zscore=apply_zscore
        )
        self.n_clusters = n_clusters
        self.random_state = random_state
        self.labels = self._compute_kmeans_labels()
        self.cluster_labels = np.unique(self.labels)
        self.timeseries = self.permute_timeseries(self.timeseries)

    def _compute_kmeans_labels(self):
        kmeans = KMeans(n_clusters=self.n_clusters, random_state=self.random_state).fit(self.timeseries.T)
        return kmeans.labels_

    def permute_timeseries(self, timeseries: np.ndarray):
        # TODO: optimize this method
        assert timeseries.shape[-1] == int(self.shape[-1])
        permuted_timeseries = np.zeros_like(timeseries)
        position = 0
        for cluster in self.cluster_labels:
            for i in range(int(self.shape[-1])):
                if self.labels[i] == cluster:
                    permuted_timeseries[:, position] = timeseries[:, i]
                    position += 1
        return permuted_timeseries


class VisualisePCA(Visualise):
    """
    Visualise the time series using PCA algorithm of dimensionality reduction. PCA is apply on the variable.

    TODO: generalise the class to work with Size and to save the results in a file.
    """

    def __init__(
            self,
            timeseries: Any,
            shape: Optional[DimensionsLike] = None,
            apply_zscore: bool = False,
            n_PC: int = 2
    ):
        """
        :param timeseries: Time series of shape (n_time_steps, n_neurons).
        :param shape: Shape of the time series. If shape is None, the shape is inferred from the time series. Useful
        to identify the dtype of the time series. If the shape is Size, make sure to set the name of the dimensions as
        you want them to be displayed.
        :param apply_zscore: Whether to apply z-score or not.
        :param n_PC: Number of principal components.
        """
        super().__init__(
            timeseries=timeseries,
            shape=shape,
            apply_zscore=apply_zscore
        )
        self.n_PC = n_PC
        self.params = {}
        self.pca_transform = None
        self.reduced_timeseries, self.params["var_ratio"], self.params["var_ratio_cumsum"] = self._compute_pca(n_PC)
        self.kmean_label = None

    def _compute_pca(self, n_PC: int):
        """
        Compute PCA of the time series.

        :return: timeseries in PCA space, variance ratio, variance ratio cumulative sum.
        """
        self.pca_transform = PCA(n_components=n_PC)
        self.pca_transform.fit(self.timeseries)
        reduced_timeseries = self.pca_transform.transform(self.timeseries)
        return reduced_timeseries, self.pca_transform.explained_variance_ratio_, self.pca_transform.explained_variance_ratio_.cumsum()

    def with_kmeans(self, n_clusters: int = 13, random_state: int = 0):
        """
        Apply K-means clustering to the PCA space as coloring.

        :param n_clusters: Number of clusters.
        :param random_state: Determines random number generation for centroid initialization.
            Example: VisualisePCA(data).with_kmeans(n_clusters=13, random_state=0).scatter_pca()
        """
        kmeans = KMeans(n_clusters=n_clusters, random_state=random_state).fit(self.reduced_timeseries)
        self.kmean_label = kmeans.labels_
        return self

    def scatter_pca(self, PCs: Tuple[int, ...] = (1, 2), color_sample: bool = False):
        """
        Plot the scatter plot of the PCA space in 2D or 3D.

        :param PCs: List of PCs to plot. Always a list of length 2 or 3
        :param color_sample: Whether to color the sample or not.
        """
        dimension = len(PCs)
        if dimension < 2 or dimension > 3:
            raise ValueError("PCs must be a tuple of 2 or 3 elements. Can only plot PCs in 2D or 3D")
        if self.kmean_label is not None and color_sample:
            raise ValueError("You can only apply color based on k-mean or the sample, not both")
        if max(PCs) > self.n_PC:
            raise ValueError("PCs must be less than or equal to the number of PC")
        color = None
        if self.kmean_label is not None:
            color = self.kmean_label
        if color_sample:
            color = range(int(self.shape[-1]))

        if dimension == 2:
            plt.title("Two-dimensional PCA embedding")
            plt.xlabel(f"PC {PCs[0]}")
            plt.ylabel(f"PC {PCs[1]}")
            plt.scatter(self.reduced_timeseries[:, PCs[0] - 1], self.reduced_timeseries[:, PCs[1] - 1],
                        c=color, cmap="RdBu_r")
            if self.kmean_label is not None or color_sample:
                plt.colorbar()
        if dimension == 3:
            fig, ax = plt.subplots(figsize=(16, 8))
            ax = fig.add_subplot(projection="3d")
            ax.set_title("Three-dimensional PCA embedding")
            ax.set_xlabel(f"PC {PCs[0]}")
            ax.set_ylabel(f"PC {PCs[1]}")
            ax.set_zlabel(f"PC {PCs[2]}")
            ax.scatter(
                self.reduced_timeseries[:, PCs[0] - 1],
                self.reduced_timeseries[:, PCs[1] - 1],
                self.reduced_timeseries[:, PCs[2] - 1],
                c=color, cmap="RdBu_r"
            )
        plt.show()

    def trajectory_pca(
            self,
            target: Optional = None,
            PCs: Tuple[int, int] = (1, 2),
            with_smooth: bool = True,
            degree: int = 5,
            condition: float = 5,
            reduction: int = 1,
            traces: str = "all",
            fig: Optional[plt.Figure] = None,
            axes: Optional[plt.Axes] = None,
            filename: Optional[str] = None,
            show: bool = True,
            **kwargs
    ):
        """
        Plot the trajectory of the PCA space in 2D.

        :param target: Target timeseries
        :param PCs: List of PCs to plot. Always a list of length 2.
        :param with_smooth: Whether to smooth the trajectory or not.
        :param degree: Degree of the polynomial used for smoothing.
        :param condition: Smoothing condition.
        :param reduction: Number by which we divide the number of samples.
        :param traces: Which traces to plot. Can be "all", "PCA_space", "PCA_wrt_time" or a list of indices.
        """

        assert (fig is None and axes is None) or (fig is not None and axes is not None)
        assert traces in ["all", "PCA_space", "PCA_wrt_time"]
        if max(PCs) > self.n_PC:
            raise ValueError("PCs must be less than or equal to the number of PC")
        if target is not None:
            assert isinstance(target, Visualise)
            target_reduced = self.pca_transform.transform(target.timeseries)

        n_plot = 3 if (traces == "all") else 1
        if fig is None or axes is None:
            fig, axes = plt.subplots(n_plot, 1, figsize=kwargs.get("figsize", (16, 8)))
        else:
            assert len(axes) == n_plot, f"axes must be a list of length {n_plot}"

        if traces == "all" or traces == "PCA_space":
            axes[0].set_title("Trajectory in PCA space")
            axes[0].set_xlabel(f"PC {PCs[0]}")
            axes[0].set_ylabel(f"PC {PCs[1]}")
            if target is not None:
                axes[0].plot(
                    target_reduced[::reduction, PCs[0] - 1],
                    target_reduced[::reduction, PCs[1] - 1],
                    label="Real time series",
                    c="tab:orange"
                )
            axes[0].plot(
                self.reduced_timeseries[::reduction, PCs[0] - 1],
                self.reduced_timeseries[::reduction, PCs[1] - 1],
                label="Predicted time series",
                c="tab:blue"
            )
            bbox = dict(boxstyle="round", fc="0.8")
            axes[0].plot(self.reduced_timeseries[0, PCs[0] - 1], self.reduced_timeseries[0, PCs[1] - 1], marker="o", color="black")
            axes[0].annotate(r"$t_0$", (self.reduced_timeseries[0, PCs[0] - 1], self.reduced_timeseries[0, PCs[1] - 1]+0.8), bbox=bbox)
            axes[0].annotate(r"$t_f$",
                             (self.reduced_timeseries[-1, PCs[0] - 1], self.reduced_timeseries[-1, PCs[1] - 1] - 1.5),
                             bbox=bbox)

            axes[0].plot(self.reduced_timeseries[-1, PCs[0] - 1],self.reduced_timeseries[-1, PCs[1] - 1], marker="o", color="black")
            if "box_aspect" in kwargs:
                axes[0].set_box_aspect(kwargs["box_aspect"])
            axes[0].legend()
            axes[0].set_xlabel(f"PC {PCs[0]}")
            axes[0].set_ylabel(f"PC {PCs[1]}")
            axes[0].set_title("Trajectory in PCA space")

        if traces == "PCA_wrt_time":
            axes[0].set_title("Trajectory in PCA space with respect to time")
            axes[0].set_xlabel("Time")
            axes[0].set_ylabel(f"PC {PCs[0]}")
            if target is not None:
                axes[0].plot(
                    target_reduced[::reduction, PCs[0] - 1].flatten(),
                    label="Real time series",
                    c="tab:orange"
                )
            axes[0].plot(
                self.reduced_timeseries[::reduction, PCs[0] - 1].flatten(),
                label="Predicted time series",
                c="tab:blue"
            )
            axes[0].legend()
            axes[0].set_xlabel("Time [-]")
            axes[0].set_ylabel(f"PC {PCs[0]}")
            if target is not None:
                pVar = PVarianceLoss()(self.reduced_timeseries[::reduction, PCs[0] - 1],
                                       target_reduced[::reduction, PCs[0] - 1])
                axes[0].set_title(f"Trajectory in PCA space with respect to time (pVar = {to_numpy(pVar).item():.4f})")
            else:
                axes[0].set_title("Trajectory in PCA space with respect to time")
            if "box_aspect" in kwargs:
                axes[0].set_box_aspect(kwargs["box_aspect"])

        if traces == "all":
            axes[n_plot - 2].set_title("Trajectory in PCA space with respect to time")
            axes[n_plot - 2].set_xlabel("Time")
            axes[n_plot - 2].set_ylabel(f"PC {PCs[0]}")
            if target is not None:
                axes[n_plot - 2].plot(
                    target_reduced[::reduction, PCs[0] - 1].flatten(),
                    label="Real timeseries",
                    c="tab:orange"
                )
            axes[n_plot - 2].plot(
                self.reduced_timeseries[::reduction, PCs[0] - 1].flatten(),
                label="Predicted timeseries",
                c="tab:blue"
            )
            axes[n_plot - 2].legend()
            axes[n_plot - 2].set_xlabel("Time [-]")
            axes[n_plot - 2].set_ylabel(f"PC {PCs[0]}")
            if target is not None:
                pVar = PVarianceLoss()(self.reduced_timeseries[::reduction, PCs[0] - 1],
                                       target_reduced[::reduction, PCs[0] - 1])
                axes[n_plot - 2].set_title(
                    f"Trajectory in PCA space with respect to time (pVar = {to_numpy(pVar).item():.4f})")
            else:
                axes[n_plot - 2].set_title("Trajectory in PCA space with respect to time")
            if "box_aspect" in kwargs:
                axes[n_plot - 1].set_box_aspect(kwargs["box_aspect"])

            axes[n_plot - 1].set_title("Trajectory in PCA space with respect to time")
            axes[n_plot - 1].set_xlabel("Time")
            axes[n_plot - 1].set_ylabel(f"PC {PCs[1]}")
            if target is not None:
                axes[n_plot - 1].plot(
                    target_reduced[::reduction, PCs[1] - 1].flatten(),
                    label="Real timeseries",
                    c="tab:orange"
                )
            axes[n_plot - 1].plot(
                self.reduced_timeseries[::reduction, PCs[1] - 1].flatten(),
                label="Predicted timeseries",
                c="tab:blue"
            )
            axes[n_plot - 1].legend()
            axes[n_plot - 1].set_xlabel("Time [-]")
            axes[n_plot - 1].set_ylabel(f"PC {PCs[1]}")
            if target is not None:
                pVar = PVarianceLoss()(self.reduced_timeseries[::reduction, PCs[1] - 1],
                                       target_reduced[::reduction, PCs[1] - 1])
                axes[n_plot - 1].set_title(
                    f"Trajectory in PCA space with respect to time (pVar = {to_numpy(pVar).item():.4f})")
            else:
                axes[n_plot - 1].set_title("Trajectory in PCA space with respect to time")
            if "box_aspect" in kwargs:
                axes[n_plot - 1].set_box_aspect(kwargs["box_aspect"])

            fig.tight_layout()
            if filename is not None:
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                fig.savefig(filename, dpi=kwargs.get("dpi", 300))
            if show:
                plt.show()
            if kwargs.get("close", False):
                plt.close(fig)
            return fig, axes

        plt.show()


class VisualiseUMAP(Visualise):
    """
    TODO: generalise the class to work with Size and to save the results in a file.
    """

    def __init__(
            self,
            timeseries: Any,
            shape: Optional[DimensionsLike] = None,
            apply_zscore: bool = False,
            n_neighbors: int = 10,
            min_dist: float = 0.5,
            n_components: int = 2,
            umap_transform: Optional[Any] = None,
            random_state: Optional[int] = None,
    ):
        super().__init__(
            timeseries=timeseries,
            shape=shape,
            apply_zscore=apply_zscore
        )
        self.n_neighbors = n_neighbors
        self.min_dist = min_dist
        self.n_components = n_components
        self.kmeans_label = None
        self.umap_transform = umap_transform
        self.random_state = random_state
        if umap_transform is None:
            self.umap_transform = self.compute_umap_transform()
        self.reduced_timeseries = self.reduce_timeseries(self.timeseries)

    def compute_umap_transform(self, data: Optional[Any] = None):
        try:
            import umap
        except ImportError:
            raise ImportError("You must install umap-learn to use this class. `pip install umap-learn`")
        if data is None:
            data = self.timeseries
        self.umap_transform = umap.UMAP(
            n_neighbors=self.n_neighbors,
            min_dist=self.min_dist,
            n_components=self.n_components,
            metric='euclidean',
            random_state=self.random_state,
        )
        self.umap_transform.fit(data)
        return self.umap_transform

    def reduce_timeseries(self, timeseries: Any):
        if self.umap_transform is None:
            self.compute_umap_transform()
        return self.umap_transform.transform(timeseries)

    def with_kmeans(self, n_clusters: int = 13, random_state: int = 0):
        """
        Apply K-means clustering to the PCA space as coloring.
        :param n_clusters: Number of clusters.
        :param random_state: Determines random number generation for centroid initialization.
            Example: VisualisePCA(data).with_kmeans(n_clusters=13, random_state=0).scatter_umap()
        """
        kmeans = KMeans(n_clusters=n_clusters, random_state=random_state).fit(self.reduced_timeseries)
        self.kmeans_label = kmeans.labels_
        return self

    def scatter_umap(self, UMAPs: Tuple[int, ...] = (1, 2), color_sample: bool = False):
        """
        Plot the scatter plot of the UMAP space in 2D or 3D.
        :param UMAPs: List of UMAPs to plot. Always a list of length 2 or 3
        :param color_sample: Whether to color the sample or not.
        """
        dimension = len(UMAPs)
        if dimension < 2 or dimension > 3:
            raise ValueError("UMAPs must be a tuple of 2 or 3 elements. Can only plot UMAPs in 2D or 3D")
        if self.kmeans_label is not None and color_sample:
            raise ValueError("You can only apply color based on k-mean or the sample, not both")
        if max(UMAPs) > self.n_components:
            raise ValueError("UMAPs must be less than or equal to the number of UMAP")
        color = None
        if self.kmeans_label is not None:
            color = self.kmeans_label
        if color_sample:
            color = range(self.num_sample)

        if dimension == 2:
            plt.title("Two-dimensional UMAP embedding")
            plt.xlabel(f"UMAP {UMAPs[0]}")
            plt.ylabel(f"UMAP {UMAPs[1]}")
            plt.scatter(
                self.reduced_timeseries[:, UMAPs[0] - 1], self.reduced_timeseries[:, UMAPs[1] - 1],
                c=color, cmap="RdBu_r"
            )
            if self.kmeans_label is not None or color_sample:
                plt.colorbar()
        if dimension == 3:
            fig, ax = plt.subplots(figsize=(16, 8))
            ax = fig.add_subplot(projection="3d")
            ax.set_title("Three-dimensional UMAP embedding")
            ax.set_xlabel(f"UMAP {UMAPs[0]}")
            ax.set_ylabel(f"UMAP {UMAPs[1]}")
            ax.set_zlabel(f"UMAP {UMAPs[2]}")
            ax.scatter(
                self.reduced_timeseries[:, UMAPs[0] - 1],
                self.reduced_timeseries[:, UMAPs[1] - 1],
                self.reduced_timeseries[:, UMAPs[2] - 1],
                c=color, cmap="RdBu_r"
            )
        plt.show()

    def trajectory_umap(
            self,
            target: Optional[Visualise] = None,
            UMAPs: Tuple = (1, 2),
            reduction: int = 1,
            traces: str = "all",
            fig: Optional[plt.Figure] = None,
            axes: Optional[plt.Axes] = None,
            filename: Optional[str] = None,
            show: bool = True,
            **kwargs
    ):
        """
        Plot the trajectory of the UMAP space in 2D.
        TODO: clean up the code

        :param target: Visualise object to plot the trajectory with.
        :param UMAPs: List of UMAPs to plot. Always a list of length 2.
        :param reduction: Number by which we divide the number of samples.
        :param traces: Which traces to plot. Can be "all", "UMAP_space" or "UMAP_wrt_time".
        :param fig: Figure to plot on.
        :param axes: Axes to plot on.
        :param filename: Filename to save the figure to.
        :param show: Whether to show the figure.
        """
        assert (fig is None and axes is None) or (fig is not None and axes is not None)
        assert traces in ["all", "UMAP_space", "UMAP_wrt_time"]
        if max(UMAPs) > self.n_components:
            raise ValueError("UMAPs must be less than or equal to the number of UMAP")
        if target is not None:
            assert isinstance(target, Visualise)
            target_reduced = self.umap_transform.transform(target.timeseries)

        n_plot = 3 if (traces == "all") else 1
        if fig is None or axes is None:
            fig, axes = plt.subplots(n_plot, 1, figsize=kwargs.get("figsize", (16, 8)))
        else:
            assert np.asarray(axes).size == n_plot, f"axes must have length {n_plot}"

        if traces == "all" or traces == "UMAP_space":
            axes[0].set_title("Trajectory in the UMAP space")
            axes[0].set_xlabel(f"UMAP {UMAPs[0]}")
            axes[0].set_ylabel(f"UMAP {UMAPs[1]}")
            if target is not None:
                axes[0].plot(
                    target_reduced[::reduction, UMAPs[0] - 1],
                    target_reduced[::reduction, UMAPs[1] - 1],
                    label="Real timeseries",
                    c="tab:orange"
                )
            axes[0].plot(
                self.reduced_timeseries[::reduction, UMAPs[0] - 1],
                self.reduced_timeseries[::reduction, UMAPs[1] - 1],
                label="Predicted timeseries",
                c="tab:blue"
            )
            bbox = dict(boxstyle="round", fc="0.8")
            axes[0].plot(
                self.reduced_timeseries[0, UMAPs[0] - 1], self.reduced_timeseries[0, UMAPs[1] - 1], marker="o",
                color="black"
            )
            axes[0].annotate(
                r"$t_0$", (self.reduced_timeseries[0, UMAPs[0] - 1], self.reduced_timeseries[0, UMAPs[1] - 1] + 2),
                bbox=bbox
            )
            axes[0].annotate(
                r"$t_f$",
                (self.reduced_timeseries[-1, UMAPs[0] - 1], self.reduced_timeseries[-1, UMAPs[1] - 1] - 3.7),
                bbox=bbox
            )

            axes[0].plot(
                self.reduced_timeseries[-1, UMAPs[0] - 1], self.reduced_timeseries[-1, UMAPs[1] - 1], marker="o",
                color="black"
            )
            if "box_aspect" in kwargs:
                axes[0].set_box_aspect(kwargs["box_aspect"])
            axes[0].legend()
            axes[0].set_xlabel(f"UMAP {UMAPs[0]}")
            axes[0].set_ylabel(f"UMAP {UMAPs[1]}")
            axes[0].set_title("Trajectory in the UMAP space")

        if traces == "UMAP_wrt_time":
            axes[0].set_title("Trajectory in the UMAP space with respect to time")
            axes[0].set_xlabel("Time")
            axes[0].set_ylabel(f"UMAP {UMAPs[0]}")
            if target is not None:
                axes[0].plot(
                    target_reduced[::reduction, UMAPs[0] - 1].flatten(),
                    label=f"Real timeseries",
                    c="tab:orange"
                )
            axes[0].plot(
                self.reduced_timeseries[::reduction, UMAPs[0] - 1].flatten(),
                label=f"Predicted timeseries",
                c="tab:blue"
            )
            axes[0].legend()
            axes[0].set_xlabel("Time [-]")
            axes[0].set_ylabel(f"UMAP {UMAPs[0]}")
            if target is not None:
                pVar = PVarianceLoss()(
                    self.reduced_timeseries[::reduction, UMAPs[0] - 1],
                    target_reduced[::reduction, UMAPs[0] - 1]
                )
                axes[0].set_title(
                    f"Trajectory in the UMAP space with respect to time (pVar={to_numpy(pVar).item():.4f})"
                )
            else:
                axes[0].set_title("Trajectory in the UMAP space with respect to time")
            if "box_aspect" in kwargs:
                axes[0].set_box_aspect(kwargs["box_aspect"])

        if traces == "all":
            axes[n_plot - 2].set_title("Trajectory in the UMAP space with respect to time")
            axes[n_plot - 2].set_xlabel("Time")
            axes[n_plot - 2].set_ylabel(f"UMAP {UMAPs[0]}")
            if target is not None:
                axes[n_plot - 2].plot(
                    target_reduced[::reduction, UMAPs[0] - 1].flatten(),
                    label=f"Real timeseries",
                    c="tab:orange"
                )
            axes[n_plot - 2].plot(
                self.reduced_timeseries[::reduction, UMAPs[0] - 1].flatten(),
                label=f"Predicted timeseries",
                c="tab:blue"
            )
            axes[n_plot - 2].legend()
            axes[n_plot - 2].set_xlabel("Time [-]")
            axes[n_plot - 2].set_ylabel(f"UMAP {UMAPs[0]}")
            if target is not None:
                pVar = PVarianceLoss()(
                    self.reduced_timeseries[::reduction, UMAPs[0] - 1],
                    target_reduced[::reduction, UMAPs[0] - 1]
                )
                axes[n_plot - 2].set_title(
                    f"Trajectory in the UMAP space with respect to time (pVar = {to_numpy(pVar).item():.3f})"
                )
            else:
                axes[n_plot - 2].set_title("Trajectory in the UMAP space with respect to time")
            if "box_aspect" in kwargs:
                axes[n_plot - 2].set_box_aspect(kwargs["box_aspect"])

            axes[n_plot - 1].set_title("Trajectory in the UMAP space with respect to time")
            axes[n_plot - 1].set_xlabel("Time")
            axes[n_plot - 1].set_ylabel(f"UMAP {UMAPs[1]}")
            if target is not None:
                axes[n_plot - 1].plot(
                    target_reduced[::reduction, UMAPs[1] - 1].flatten(),
                    label=f"Real timeseries",
                    c="tab:orange"
                )
            axes[n_plot - 1].plot(
                self.reduced_timeseries[::reduction, UMAPs[1] - 1].flatten(),
                label=f"Predicted timeseries",
                c="tab:blue"
            )
            axes[n_plot - 1].legend()
            axes[n_plot - 1].set_xlabel("Time [-]")
            axes[n_plot - 1].set_ylabel(f"UMAP {UMAPs[1]}")
            if target is not None:
                pVar = PVarianceLoss()(
                    self.reduced_timeseries[::reduction, UMAPs[1] - 1],
                    target_reduced[::reduction, UMAPs[1] - 1]
                )
                axes[n_plot - 1].set_title(
                    f"Trajectory in the UMAP space with respect to time (pVar = {to_numpy(pVar).item():.3f})"
                )
            else:
                axes[n_plot - 1].set_title("Trajectory in the UMAP space with respect to time")
            if "box_aspect" in kwargs:
                axes[n_plot - 1].set_box_aspect(kwargs["box_aspect"])

        fig.tight_layout()
        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            fig.savefig(filename, dpi=kwargs.get("dpi", 300))
        if show:
            plt.show()
        if kwargs.get("close", False):
            plt.close(fig)
        return fig, axes


class VisualiseDBSCAN(Visualise):
    """
    TODO: generalise the class to work with Size and to save the results in a file.
    """

    def __init__(
            self,
            timeseries: Any,
            shape: Optional[DimensionsLike] = None,
            apply_zscore: bool = True,
            eps: float = 25,
            min_samples: int = 3,
    ):
        super().__init__(
            timeseries=timeseries,
            shape=shape,
            apply_zscore=apply_zscore
        )
        self.eps = eps
        self.min_samples = min_samples
        self.timeseries = self._permute_timeseries()

    def _compute_dbscan(self):
        dbscan = DBSCAN(eps=self.eps, min_samples=self.min_samples).fit(self.timeseries)
        print(dbscan.labels_)
        return dbscan.labels_

    def _permute_timeseries(self):
        labels = self._compute_dbscan()
        cluster_labels = np.unique(labels)
        permuted_timeseries = np.zeros_like(self.timeseries)
        position = 0
        for cluster in cluster_labels:
            for i in range(self.num_sample):
                if labels[i] == cluster:
                    permuted_timeseries[i] = self.timeseries[position]
                    position += 1
        return permuted_timeseries



import os
from typing import Iterable, Optional, Tuple, Any

import numpy as np
import torch
from matplotlib import pyplot as plt


@torch.no_grad()
def visualize_init_final_weights(
        initial_wights: Any,
        final_weights: Any,
        filename: Optional[str] = None,
        show: bool = False,
        fig: Optional[plt.Figure] = None,
        axes: Optional[Iterable[plt.Axes]] = None,
        **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    import neurotorch as nt
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    from matplotlib import colors
    initial_wights, final_weights = nt.to_numpy(initial_wights), nt.to_numpy(final_weights)

    assert (fig is None) == (axes is None), "If fig is None, axes must be None"
    if fig is None:
        fig, axes = plt.subplots(1, 2, figsize=(12, 8))
    else:
        assert isinstance(axes, Iterable), "If fig is not None, axes must be an iterable"
    axes_view = np.asarray(axes).ravel()

    if kwargs.get("cmp_zero_center", True):
        min_value = min(np.min(initial_wights), np.min(final_weights))
        max_value = max(np.max(initial_wights), np.max(final_weights))
        if not np.isclose(min_value, max_value) and min_value < max_value and min_value < 0 < max_value:
            divnorm = colors.TwoSlopeNorm(
                vmin=min_value,
                vcenter=0.0,
                vmax=max_value
            )
            im = axes_view[0].imshow(initial_wights, cmap="RdBu_r", norm=divnorm)
        else:
            im = axes_view[0].imshow(initial_wights, cmap="RdBu_r")
    else:
        im = axes_view[0].imshow(initial_wights, cmap="RdBu_r")
    axes_view[1].imshow(final_weights, cmap=im.cmap, extent=im.get_extent())
    cax = inset_axes(
        axes_view[1],
        width="5%",
        height="100%",
        loc='lower left',
        bbox_to_anchor=(1.05, 0.0, 1.0, 1.0),
        bbox_transform=axes_view[1].transAxes,
        borderpad=0,
    )

    if kwargs.get('compute_dale', "dale_law_kwargs" in kwargs):
        dale_law_kwargs = kwargs.get('dale_law_kwargs', {})
        dale_law_kwargs.setdefault('seed', 0)
        initial_dale = nt.to_numpy(nt.DaleLaw([nt.to_tensor(initial_wights)], **dale_law_kwargs)())
        final_dale = nt.to_numpy(nt.DaleLaw([nt.to_tensor(final_weights)], **dale_law_kwargs)())
        axes_view[0].set_title(f"Initial Weights (Dale loss = {initial_dale:.3f})")
        axes_view[1].set_title(f"Final Weights (Dale loss = {final_dale:.3f})")
    else:
        axes_view[0].set_title("Initial Weights")
        axes_view[1].set_title("Final Weights")

    fig.colorbar(im, cax=cax)
    fig.set_tight_layout(False)
    if filename is not None:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        fig.savefig(filename)
    if show:
        plt.show()
    return fig, axes





from typing import Optional, Tuple, Union

import torch

from .base import BaseNeuronsLayer
from ...dimension import SizeTypes
from ...transforms import to_tensor


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseNeuronsLayer])
class WilsonCowanLayer(BaseNeuronsLayer):
    """
    This layer is use for Wilson-Cowan neuronal dynamics.
    This dynamic is also referred to as firing rate model.
    Wilson-Cowan dynamic is great for neuronal calcium activity.
    This layer use recurrent neural network (RNN).
    The number of parameters that are trained is N^2 (+2N if mu and r is train)
    where N is the number of neurons.

    For references, please read:

        - Excitatory and Inhibitory Interactions in Localized Populations of Model Neurons :cite:t:`wilson1972excitatory`
        - Beyond Wilson-Cowan dynamics: oscillations and chaos without inhibitions :cite:t:`PainchaudDoyonDesrosiers2022`
        - Neural Network dynamic :cite:t:`VogelsTimRajanAbbott2005NeuralNetworkDynamics`.

    The Wilson-Cowan dynamic is one of many dynamical models that can be used
    to model neuronal activity. To explore more continuous and Non-linear dynamics,
    please read Nonlinear Neural Network: Principles, Mechanisms, and Architecture :cite:t:`GROSSBERG198817`.


    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            dt: float = 1e-3,
            use_recurrent_connection: bool = False,
            device=None,
            **kwargs
    ):
        """
        :param input_size: size of the input
        :type input_size: Optional[SizeTypes]
        :param output_size: size of the output
            If we are predicting time series -> input_size = output_size
        :type output_size: Optional[SizeTypes]
        :param learning_type: Type of learning for the gradient descent
        :type learning_type: LearningType
        :param dt: Time step (Euler's discretisation)
        :type dt: float
        :param device: device for computation
        :type device: torch.device
        :param kwargs: Additional parameters for the Wilson-Cowan dynamic.

        :keyword Union[torch.Tensor, np.ndarray] forward_weights: Forward weights of the layer.
        :keyword float std_weight: Instability of the initial random matrix.
        :keyword Union[float, torch.Tensor] mu: Activation threshold. If torch.Tensor -> shape (1, number of neurons).
        :keyword float mean_mu: Mean of the activation threshold (if learn_mu is True).
        :keyword float std_mu: Standard deviation of the activation threshold (if learn_mu is True).
        :keyword bool learn_mu: Whether to train the activation threshold.
        :keyword float tau: Decay constant of RNN unit.
        :keyword bool learn_tau: Whether to train the decay constant.
        :keyword float r: Transition rate of the RNN unit. If torch.Tensor -> shape (1, number of neurons).
        :keyword float mean_r: Mean of the transition rate (if learn_r is True).
        :keyword float std_r: Standard deviation of the transition rate (if learn_r is True).
        :keyword bool learn_r: Whether to train the transition rate.

        Remarks: Parameter mu and r can only be a parameter as a vector.
        """
        super(WilsonCowanLayer, self).__init__(
            input_size=input_size,
            output_size=output_size,
            use_recurrent_connection=use_recurrent_connection,
            dt=dt,
            device=device,
            **kwargs
        )
        self.std_weight = self.kwargs["std_weight"]
        self.mu = torch.nn.Parameter(to_tensor(self.kwargs["mu"]).to(self.device), requires_grad=False)
        self.mean_mu = self.kwargs["mean_mu"]
        self.std_mu = self.kwargs["std_mu"]
        self.learn_mu = self.kwargs["learn_mu"]
        self.tau_sqrt = torch.nn.Parameter(
            torch.sqrt(to_tensor(self.kwargs["tau"])).to(self.device), requires_grad=False
        )
        self.learn_tau = self.kwargs["learn_tau"]
        self.r_sqrt = torch.nn.Parameter(
            torch.sqrt(to_tensor(self.kwargs["r"], dtype=torch.float32)).to(self.device), requires_grad=False
        )
        self.mean_r = self.kwargs["mean_r"]
        self.std_r = self.kwargs["std_r"]
        self.learn_r = self.kwargs["learn_r"]
        self.activation = self._init_activation(self.kwargs["activation"])

    def _set_default_kwargs(self):
        self.kwargs.setdefault("std_weight", 1.0)
        self.kwargs.setdefault("mu", 0.0)
        self.kwargs.setdefault("tau", 1.0)
        self.kwargs.setdefault("learn_tau", False)
        self.kwargs.setdefault("learn_mu", False)
        self.kwargs.setdefault("mean_mu", 2.0)
        self.kwargs.setdefault("std_mu", 0.0)
        self.kwargs.setdefault("r", 0.0)
        self.kwargs.setdefault("learn_r", False)
        self.kwargs.setdefault("mean_r", 2.0)
        self.kwargs.setdefault("std_r", 0.0)
        self.kwargs.setdefault("hh_init", "inputs")
        self.kwargs.setdefault("activation", "sigmoid")

    def _assert_kwargs(self):
        assert self.std_weight >= 0.0, "std_weight must be greater or equal to 0.0"
        assert self.std_mu >= 0.0, "std_mu must be greater or equal to 0.0"
        assert self.tau > self.dt, "tau must be greater than dt"

    @property
    def r(self):
        """
        This property is used to ensure that the transition rate will never be negative if trained.
        """
        return torch.pow(self.r_sqrt, 2)

    @r.setter
    def r(self, value):
        self.r_sqrt.data = torch.sqrt(torch.abs(to_tensor(value, dtype=torch.float32))).to(self.device)

    @property
    def tau(self):
        """
        This property is used to ensure that the decay constant will never be negative if trained.
        """
        return torch.pow(self.tau_sqrt, 2)

    @tau.setter
    def tau(self, value):
        self.tau_sqrt.data = torch.sqrt(torch.abs(to_tensor(value, dtype=torch.float32))).to(self.device)

    def initialize_weights_(self):
        """
        Initialize the parameters (weights) that will be trained.
        """
        super().initialize_weights_()
        if self.kwargs.get("forward_weights", None) is not None:
            self._forward_weights.data = to_tensor(self.kwargs["forward_weights"]).to(self.device)
        else:
            torch.nn.init.normal_(self._forward_weights, mean=0.0, std=self.std_weight)

        # If mu is not a parameter, it takes the value 0.0 unless stated otherwise by user
        # If mu is a parameter, it is initialized as a vector with the correct mean and std
        # unless stated otherwise by user.
        if self.learn_mu:
            if self.mu.dim() == 0:  # if mu is a scalar and a parameter -> convert it to a vector
                self.mu.data = torch.empty((1, int(self.output_size)), dtype=torch.float32, device=self.device)
            self.mu = torch.nn.Parameter(self.mu, requires_grad=self.requires_grad)
            torch.nn.init.normal_(self.mu, mean=self.mean_mu, std=self.std_mu)
        if self.learn_r:
            _r = torch.empty((1, int(self.output_size)), dtype=torch.float32, device=self.device)
            torch.nn.init.normal_(_r, mean=self.mean_r, std=self.std_r)
            self.r_sqrt = torch.nn.Parameter(torch.sqrt(torch.abs(_r)), requires_grad=self.requires_grad)
        if self.learn_tau:
            self.tau_sqrt = torch.nn.Parameter(self.tau_sqrt, requires_grad=self.requires_grad)

    def create_empty_state(self, batch_size: int = 1, **kwargs) -> Tuple[torch.Tensor]:
        if self.kwargs["hh_init"] == "zeros":
            state = [torch.zeros(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
            ) for _ in range(1)]
        elif self.kwargs["hh_init"] == "random":
            mu, std = self.kwargs.get("hh_init_mu", 0.0), self.kwargs.get("hh_init_std", 1.0)
            gen = torch.Generator(device=self.device)
            gen.manual_seed(self.kwargs.get("hh_init_seed", 0))
            state = [(torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            ) * std + mu) for _ in range(1)]
        elif self.kwargs["hh_init"] == "inputs":
            assert "inputs" in kwargs, "inputs must be provided to initialize the state"
            assert kwargs["inputs"].shape == (batch_size, int(self.output_size))
            state = (kwargs["inputs"].clone(),)
        else:
            state = super().create_empty_state(batch_size, **kwargs)
        return tuple(state)

    def forward(
            self,
            inputs: torch.Tensor,
            state: Optional[Tuple[torch.Tensor, ...]] = None,
            **kwargs
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor]]:
        """
        Forward pass.
        With Euler discretisation, Wilson-Cowan equation becomes:

        output = input * (1 - dt/tau) + dt/tau * (1 - input @ r) * sigmoid(input @ forward_weight - mu)

        :param inputs: time series at a time t of shape (batch_size, number of neurons)
            Remark: if you use to compute a time series, use batch_size = 1.
        :type inputs: torch.Tensor
        :param state: State of the layer (only for SNN -> not use for RNN)
        :type state: Optional[Tuple[torch.Tensor, ...]]

        :return: (time series at a time t+1, State of the layer -> None)
        :rtype: Tuple[torch.Tensor, Tuple[torch.Tensor, ...]]
        """
        batch_size, nb_features = inputs.shape
        hh, = self._init_forward_state(state, batch_size, inputs=inputs)
        ratio_dt_tau = self.dt / self.tau

        if self.use_recurrent_connection:
            rec_inputs = torch.matmul(hh, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_inputs = 0.0

        transition_rate = (1 - hh * self.r)
        activation = self.activation(rec_inputs + torch.matmul(inputs, self.forward_weights) - self.mu)
        output = hh * (1 - ratio_dt_tau) + transition_rate * activation * ratio_dt_tau
        return output, (torch.clone(output),)


class WilsonCowanCURBDLayer(WilsonCowanLayer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create_empty_state(self, batch_size: int = 1, **kwargs) -> Tuple[torch.Tensor, ...]:
        if self.kwargs["hh_init"].lower() == "given":
            assert "h0" in self.kwargs, "h0 must be provided as a tuple of tensors when hh_init is 'given'."
            h0 = self.kwargs["h0"]
            assert isinstance(h0, (tuple, list)), "h0 must be a tuple of tensors."
            state = [to_tensor(h0_, dtype=torch.float32).to(self.device) for h0_ in h0]
        else:
            state = super().create_empty_state(batch_size, **kwargs)
        return tuple(state)

    def forward(
            self,
            inputs: torch.Tensor,
            state: Optional[Tuple[torch.Tensor, ...]] = None,
            **kwargs
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, ...]]:
        batch_size, nb_features = inputs.shape

        out_shape = tuple(inputs.shape[:-1]) + (self.forward_weights.shape[-1],)  # [*, f_out]
        inputs_view = inputs.view(-1, inputs.shape[-1])  # [*, f_in] -> [B, f_in]

        hh, = self._init_forward_state(state, batch_size, inputs=inputs_view, **kwargs)  # [B, f_out]
        post_activation = self.activation(hh)  # [B, f_out]

        if self.use_recurrent_connection:
            # [B, f_out] @ [f_out, f_out] -> [B, f_out]
            rec_inputs = torch.matmul(post_activation, torch.mul(self.recurrent_weights, self.rec_mask))
        else:
            rec_inputs = 0.0

        # [B, f_in] @ [f_in, f_out] -> [B, f_out]
        weighted_current = torch.matmul(inputs_view, self.forward_weights)
        jr = (rec_inputs + weighted_current)  # [B, f_out]
        next_hh = hh + self.dt * (-hh + jr) / self.tau  # [B, f_out]
        output = post_activation.view(out_shape)  # [B, f_out] -> [*, f_out]
        return output, (next_hh,)




from typing import Optional, Tuple

import torch

from neurotorch.dimension import SizeTypes
from neurotorch.modules import BaseLayer


class DimensionsCat:
    def __init__(self, start_axis: int, end_axis: int = -1):
        self.start_axis = start_axis
        self.end_axis = end_axis
        assert self.start_axis < self.end_axis, \
            f"Start axis ({self.start_axis}) must be lower than end axis ({self.end_axis})."
        assert all(axis > 0 for axis in self.axes), \
            "Each axis must be greater than 0. The dimension 0 is the batch dimension and can't be concatenate " \
            "with others."

    @property
    def axes(self):
        return [self.start_axis, self.end_axis]

    def __call__(self, inputs: torch.Tensor):
        assert all(axis < inputs.shape[-1] for axis in self.axes), \
            f"Axes ({self.axes}) are not dimension of inputs ({inputs.shape})."
        return torch.flatten(inputs, start_dim=self.start_axis, end_dim=self.end_axis)











from collections import defaultdict
from typing import Any, Optional, Dict, Callable, Tuple, Union

import numpy as np
import torch
from torch.nn.functional import mse_loss
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from sklearn import metrics as sk_metrics

from .base import BaseMetrics
from ..modules import BaseModel
from ..transforms import to_tensor, to_numpy


class RegressionMetrics(BaseMetrics):
    EPSILON = 1e-12

    @staticmethod
    def get_all_metrics_names_to_func() -> Dict[str, Callable]:
        sep = RegressionMetrics.METRICS_NAMES_SEP
        return {
            f"mae{sep}mean_absolute_error": RegressionMetrics.mean_absolute_error,
            f"mse{sep}mean_squared_error": RegressionMetrics.mean_squared_error,
            f"r2": RegressionMetrics.r2,
            f"d2{sep}d2_tweedie": RegressionMetrics.d2_tweedie,
            f"p_var": RegressionMetrics.p_var,
        }

    @staticmethod
    def compute_y_true_y_pred(
            model: BaseModel,
            dataloader: DataLoader,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]], Tuple[np.ndarray, np.ndarray]]:
        if device is not None:
            model.to(device)
        model.eval()
        predictions = defaultdict(list)
        targets = defaultdict(list)
        with torch.no_grad():
            for i, (x, y_true) in tqdm(
                    enumerate(dataloader), total=len(dataloader),
                    desc=desc, disable=not verbose, position=p_bar_position,
            ):
                x = x.to(model.device)
                y_true = y_true.to(model.device)
                preds = model.get_prediction_trace(x)
                if isinstance(preds, dict):
                    if not isinstance(targets, dict):
                        targets = {k: targets for k in preds}
                    for k, v in preds.items():
                        predictions[k].extend(to_numpy(v))
                        targets[k].extend(to_numpy(y_true[k]))
                else:
                    predictions["__all__"].extend(to_numpy(preds))
                    targets["__all__"].extend(to_numpy(y_true))
        predictions = {k: np.asarray(v) for k, v in predictions.items()}
        targets = {k: np.asarray(v) for k, v in targets.items()}
        if len(targets) == 1:
            return targets[list(targets.keys())[0]], predictions[list(predictions.keys())[0]]
        return targets, predictions

    @staticmethod
    def mean_absolute_error(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
                model=model, dataloader=dataloader, device=device,
                verbose=verbose, desc=desc, p_bar_position=p_bar_position
            )

        if isinstance(y_true, dict):
            return {k: sk_metrics.mean_absolute_error(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
        return sk_metrics.mean_absolute_error(y_true.flatten(), y_pred.flatten())

    @staticmethod
    def mean_squared_error(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            return {k: sk_metrics.mean_squared_error(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
        return sk_metrics.mean_squared_error(y_true.flatten(), y_pred.flatten())

    @staticmethod
    def r2(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            return {k: sk_metrics.r2_score(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
        return sk_metrics.r2_score(y_true.flatten(), y_pred.flatten())

    @staticmethod
    def d2_tweedie(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            return {k: sk_metrics.d2_tweedie_score(y_true[k].flatten(), y_pred[k].flatten()) for k in y_true}
        return sk_metrics.d2_tweedie_score(y_true.flatten(), y_pred.flatten())

    @staticmethod
    def compute_p_var(
            y_true,
            y_pred,
            device: Optional[torch.device] = None,
            reduction: str = 'mean',
    ) -> torch.Tensor:
        y_true, y_pred = to_tensor(y_true), to_tensor(y_pred)
        if device is None:
            device = y_pred.device
        y_true, y_pred = y_true.to(device), y_pred.to(device)
        dims = list(range(y_pred.ndim))[1:]
        mse = torch.mean(mse_loss(y_pred, y_true, reduction='none'), dim=dims)
        var = torch.var(y_true, dim=dims)
        p_var_values = 1 - (mse / (var + RegressionMetrics.EPSILON))
        if reduction.lower() == 'none':
            p_var_value = p_var_values
        elif reduction.lower() == 'mean':
            p_var_value = torch.mean(p_var_values)
        elif reduction.lower() == 'sum':
            p_var_value = torch.sum(p_var_values)
        else:
            raise ValueError(f"Reduction {reduction} not recognized. Try 'none'|'sum'|'mean'.")
        return p_var_value

    @staticmethod
    def p_var(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        """
        Compute the p-value of the variance of the prediction.
        :param model: The model to use for computing the p-value.
        :param dataloader: The dataloader to use for computing the p-value.
        :param y_true: The true values.
        :param y_pred: The predicted values.
        :param device: The device to use for computing the p-value.
        :param verbose: Whether to print progress.
        :param desc: The description to use for the progress bar.
        :param p_bar_position: The position of the progress bar.
        :return: The p-value of the variance of the prediction.
        """
        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
                model=model, dataloader=dataloader, device=device,
                verbose=verbose, desc=desc, p_bar_position=p_bar_position
            )

        if isinstance(y_true, dict):
            return {
                k: to_numpy(RegressionMetrics.compute_p_var(y_true[k], y_pred[k], device))
                for k in y_true
            }
        return to_numpy(RegressionMetrics.compute_p_var(y_true, y_pred, device))

    def __call__(
            self,
            data_loader: DataLoader,
            verbose: Union[bool, int] = False
    ) -> Dict[str, Any]:
        """
        Compute the metrics for the given data_loader.
        :param data_loader: The data loader to use to compute the metrics.
        :param verbose: 0: no progress bar, 1: single progress bar, 2: progress bar for each metrics
                        True: verbose is set to 1, False: verbose is set to 0.
        :return: The metrics computed.
        """
        if isinstance(verbose, bool):
            verbose = 1 if verbose else 0
        assert verbose in [0, 1, 2], "verbose must be 0, 1 or 2"
        output = {}
        p_bar = tqdm(
            enumerate(self.metrics_functions.items()),
            total=len(self.metrics_functions),
            disable=verbose == 0,
            unit="metric",
            position=0,
        )
        self.model.eval()
        y_true, y_pred = RegressionMetrics.compute_y_true_y_pred(
            model=self.model,
            dataloader=data_loader,
            device=self.device,
            verbose=verbose == 2,
            desc="compute_y_true_y_preds",
            p_bar_position=1
        )
        for i, (metric_name, metric_func) in p_bar:
            output[metric_name] = metric_func(
                y_true=y_true,
                y_pred=y_pred,
                device=self.device,
                verbose=verbose == 2,
                desc=metric_name,
                p_bar_position=i + 2
            )
        return output



from typing import Union, Optional

import numpy as np
import torch
import torch.nn.functional

from .base import to_tensor


class MeanConv(torch.nn.Module):
    """
    Apply a weighted mean filter to the input tensor.

    """
    def __init__(
            self,
            kernel_size: int,
            alpha: Union[float, torch.Tensor, np.ndarray] = 1.0,
            learn_alpha: bool = True,
            learn_kernel: bool = True,
            activation: torch.nn.Module = torch.nn.Identity(),
            pad_value: Optional[float] = 0.0,
            pad_mode: str = "left",
    ):
        """
        Constructor for MeanConv class.

        :param kernel_size: Size of the convolution kernel.
        :param alpha: Multiplicative factor for the convolution kernel.
        :param learn_alpha: Whether to learn the alpha parameter.
        :param learn_kernel: Whether to learn the convolution kernel.
        :param activation: Activation function to apply to the output.
        :param pad_value: Value to use for padding the input tensor. This is used to ensure that the input tensor has
            a valid shape for the convolution operation. If None, the input tensor is not padded.
        :param pad_mode: Padding mode. If "left", the padding is applied to the left of the input tensor. If "right",
            the padding is applied to the right of the input tensor. If "both", the padding is applied to both sides
            of the input tensor.
        """
        super(MeanConv, self).__init__()
        self.kernel_size = kernel_size
        self.learn_kernel = learn_kernel
        self.kernel = torch.nn.Parameter(torch.ones(1, self.kernel_size, 1), requires_grad=learn_kernel)
        self.learn_alpha = learn_alpha
        self.alpha = torch.nn.Parameter(to_tensor(alpha), requires_grad=learn_alpha)
        self.activation = activation
        self.pad_value = pad_value
        self.pad_mode = pad_mode

    def extra_repr(self):
        extra_repr = f"kernel(size={self.kernel_size}, learn={self.learn_kernel})"
        if self.learn_alpha:
            extra_repr += f"\nalpha(size={tuple(self.alpha.shape)} learn={self.learn_alpha})"
        else:
            if torch.numel(self.alpha) == 1:
                extra_repr += f"\nalpha(value={self.alpha.detach().cpu().item()})"
            else:
                extra_repr += f"\nalpha(size={tuple(self.alpha.shape)})"
        return extra_repr

    def forward(self, inputs: torch.Tensor):
        batch_size, time_steps, n_units = inputs.shape
        n_pad = (self.kernel_size - (time_steps % self.kernel_size)) % self.kernel_size
        if self.pad_value is not None:
            if self.pad_mode == "left":
                pad = (0, 0, n_pad, 0)
            elif self.pad_mode == "right":
                pad = (0, 0, 0, n_pad)
            elif self.pad_mode == "both":
                pad = (0, 0, int(n_pad - n_pad // 2), int(n_pad // 2))
            else:
                raise ValueError(f"Invalid padding mode: {self.pad_mode}")
            inputs = torch.nn.functional.pad(inputs, pad=pad, value=self.pad_value)

        inputs_view = torch.reshape(inputs, (batch_size, -1, self.kernel_size, n_units))
        inputs_mean = self.alpha * torch.sum(self.kernel * inputs_view, dim=2) / self.kernel_size
        return self.activation(inputs_mean)
	








from typing import Union, Dict

import torch


def format_pred_batch(
        raw_pred_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
        y_batch: Union[torch.Tensor, Dict[str, torch.Tensor]],
):
    """
    This function format the raw pred batch to the same format as y_batch. For example, if y_batch is a dict, then
    raw_pred_batch will be converted to a dict. If raw_pred_batch is a tuple or a list, then raw_pred_batch is
    consider to be in the following format : pred, hidden_state. In this case, only pred will be taken.

    :param raw_pred_batch:
    :param y_batch:
    :return:
    """
    from .collections import maybe_unpack_singleton_dict

    if isinstance(raw_pred_batch, (tuple, list)):
        pred_batch = raw_pred_batch[0]
    elif isinstance(raw_pred_batch, (torch.Tensor, dict)):
        pred_batch = raw_pred_batch
    else:
        raise ValueError(f"Unsupported output type: {type(raw_pred_batch)}")

    if isinstance(y_batch, dict):
        if len(y_batch) == 1 and not isinstance(pred_batch, dict):
            pred_batch = {k: pred_batch for k in y_batch}
        assert isinstance(pred_batch, dict) and isinstance(y_batch, dict), \
            "If y_batch is a dict, pred must be a dict too."
        assert set(pred_batch.keys()) == set(y_batch.keys()), \
            "Keys of y_batch and pred_batch must be the same."
    else:
        if isinstance(pred_batch, dict):
            assert len(pred_batch) == 1, \
                "pred_batch must have only one key if y_batch is not a dict."
            pred_batch = maybe_unpack_singleton_dict(pred_batch)
    return pred_batch



from collections import defaultdict
from typing import Any, Optional, Dict, Callable, Tuple, Union

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from .base import BaseMetrics
from ..modules import BaseModel


class ClassificationMetrics(BaseMetrics):
    """
    This class is used to compute metrics for a classification model.
    """

    @staticmethod
    def get_all_metrics_names_to_func() -> Dict[str, Callable]:
        sep = ClassificationMetrics.METRICS_NAMES_SEP
        return {
            f"accuracy{sep}acc": ClassificationMetrics.accuracy,
            f"precision": ClassificationMetrics.precision,
            f"recall": ClassificationMetrics.recall,
            f"f1": ClassificationMetrics.f1,
            # f"auc": ClassificationMetrics.auc,
        }

    @staticmethod
    def compute_y_true_y_pred(
            model: BaseModel,
            dataloader: DataLoader,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray]], Tuple[np.ndarray, np.ndarray]]:
        if device is not None:
            model.to(device)
        model.eval()
        predictions = defaultdict(list)
        targets = defaultdict(list)
        with torch.no_grad():
            for i, (inputs, classes) in tqdm(
                    enumerate(dataloader), total=len(dataloader),
                    desc=desc, disable=not verbose, position=p_bar_position,
            ):
                if inputs.is_sparse:
                    inputs = inputs.to_dense()
                if classes.is_sparse:
                    classes = classes.to_dense()
                inputs = inputs.to(model.device)
                classes = classes.to(model.device)
                outputs = model.get_prediction_proba(inputs, re_outputs_trace=False, re_hidden_states=False)
                if isinstance(outputs, dict):
                    if not isinstance(classes, dict):
                        classes = {k: classes for k in outputs}
                    for k, v in outputs.items():
                        _, preds = torch.max(v, -1)
                        preds = preds.cpu().numpy()
                        predictions[k].extend(preds)
                        if len(classes[k].shape) > 1:
                            classes[k] = torch.argmax(classes[k], -1).view(-1)
                        targets[k].extend(classes[k].cpu().numpy())
                else:
                    _, preds = torch.max(outputs, -1)
                    preds = preds.cpu().numpy()
                    if len(classes.shape) > 1:
                        classes = torch.argmax(classes, -1).view(-1)
                    predictions["__all__"].extend(preds)
                    targets["__all__"].extend(classes.cpu().numpy())
        predictions = {k: np.asarray(v) for k, v in predictions.items()}
        targets = {k: np.asarray(v) for k, v in targets.items()}
        if len(targets) == 1:
            return targets[list(targets.keys())[0]], predictions[list(predictions.keys())[0]]
        return targets, predictions

    @staticmethod
    def accuracy(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        from sklearn.metrics import accuracy_score
        assert (y_true is None) == (y_pred is None)

        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            return {k: accuracy_score(y_true[k], y_pred[k]) for k in y_true}
        return accuracy_score(y_true, y_pred)

    @staticmethod
    def precision(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        from sklearn.metrics import precision_score
        assert (y_true is None) == (y_pred is None)

        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )
        if isinstance(y_true, dict):
            average = {k: ("macro" if len(set(v)) > 2 else "binary") for k, v in y_true.items()}
        else:
            average = "macro" if len(set(y_true)) > 2 else "binary"
        if isinstance(y_true, dict):
            return {k: precision_score(y_true[k], y_pred[k], average=average, zero_division=0) for k in y_true}
        return precision_score(y_true, y_pred, average=average, zero_division=0)

    @staticmethod
    def confusion_matrix(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[Dict[str, np.ndarray], np.ndarray]:
        from sklearn.metrics import confusion_matrix
        assert (y_true is None) == (y_pred is None)

        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            return {k: confusion_matrix(y_true[k], y_pred[k]) for k in y_true}
        return confusion_matrix(y_true, y_pred)

    @staticmethod
    def f1(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        from sklearn.metrics import f1_score
        assert (y_true is None) == (y_pred is None)

        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            average = {k: ("macro" if len(set(v)) > 2 else "binary") for k, v in y_true.items()}
        else:
            average = "macro" if len(set(y_true)) > 2 else "binary"

        if isinstance(y_true, dict):
            return {k: f1_score(y_true[k], y_pred[k], average=average, zero_division=0) for k in y_true}
        return f1_score(y_true, y_pred, average=average, zero_division=0)

    @staticmethod
    def recall(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        from sklearn import metrics as sk_metrics
        assert (y_true is None) == (y_pred is None)

        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            average = {k: ("macro" if len(set(v)) > 2 else "binary") for k, v in y_true.items()}
        else:
            average = "macro" if len(set(y_true)) > 2 else "binary"

        if isinstance(y_true, dict):
            return {k: sk_metrics.recall_score(y_true[k], y_pred[k], average=average, zero_division=0) for k in y_true}
        return sk_metrics.recall_score(y_true, y_pred, average=average, zero_division=0)

    @staticmethod
    def auc(
            model: Optional[BaseModel] = None,
            dataloader: Optional[DataLoader] = None,
            y_true: Optional[np.ndarray] = None,
            y_pred: Optional[np.ndarray] = None,
            device: Optional[torch.device] = None,
            verbose: bool = False,
            desc: Optional[str] = None,
            p_bar_position: int = 0,
    ) -> Union[float, Dict[str, float]]:
        from sklearn import metrics as sk_metrics
        assert (y_true is None) == (y_pred is None)

        if y_true is None:
            assert y_pred is None
            assert model is not None, "Either model or y_pred and y_true must be supplied."
            assert dataloader is not None, "Either model or y_pred and y_true must be supplied."
            y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
                model, dataloader, device, verbose, desc, p_bar_position
            )

        if isinstance(y_true, dict):
            rocs = {k: sk_metrics.roc_curve(y_true[k], y_pred[k]) for k in y_true}
            return {k: sk_metrics.auc(rocs[k][0], rocs[k][1]) for k in rocs}

        fpr, tpr, thresholds = sk_metrics.roc_curve(y_true, y_pred)
        return sk_metrics.auc(fpr, tpr)

    def __call__(
            self,
            data_loader: DataLoader,
            verbose: Union[bool, int] = False
    ) -> Dict[str, Any]:
        """
        Compute the metrics for the given data_loader.
        :param data_loader: The data loader to use to compute the metrics.
        :param verbose: 0: no progress bar, 1: single progress bar, 2: progress bar for each metrics
                        True: verbose is set to 1, False: verbose is set to 0.
        :return: The metrics computed.
        """
        if isinstance(verbose, bool):
            verbose = 1 if verbose else 0
        assert verbose in [0, 1, 2], "verbose must be 0, 1 or 2"
        output = {}
        p_bar = tqdm(
            enumerate(self.metrics_functions.items()),
            total=len(self.metrics_functions),
            disable=verbose == 0,
            unit="metric",
            position=0,
        )
        self.model.eval()
        y_true, y_pred = ClassificationMetrics.compute_y_true_y_pred(
            model=self.model,
            dataloader=data_loader,
            device=self.device,
            verbose=verbose == 2,
            desc="compute_y_true_y_preds",
            p_bar_position=1
        )
        for i, (metric_name, metric_func) in p_bar:
            output[metric_name] = metric_func(
                y_true=y_true,
                y_pred=y_pred,
                device=self.device,
                verbose=verbose == 2,
                desc=metric_name,
                p_bar_position=i + 2
            )
        return output



from typing import Optional, Sequence, Union, Dict, Callable, List

import torch

from .learning_algorithm import LearningAlgorithm
from ..utils import list_insert_replace_at
from ..utils.formatting import format_pred_batch


class BPTT(LearningAlgorithm):
    r"""
    Apply the backpropagation through time algorithm to the given model.
    """
    CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: str = "optimizer_state_dict"
    OPTIMIZER_PARAMS_GROUP_IDX = 0
    DEFAULT_OPTIMIZER_CLS = torch.optim.AdamW

    def __init__(
            self,
            *,
            params: Optional[Sequence[torch.nn.Parameter]] = None,
            layers: Optional[Union[Sequence[torch.nn.Module], torch.nn.Module]] = None,
            optimizer: Optional[torch.optim.Optimizer] = None,
            criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]] = None,
            **kwargs
    ):
        """
        Constructor for BPTT class.

        :param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
        :type params: Optional[Sequence[torch.nn.Parameter]]
        :param optimizer: The optimizer to use. If not provided, torch.optim.Adam is used.
        :type optimizer: Optional[torch.optim.Optimizer]
        :param criterion: The criterion to use. If not provided, torch.nn.MSELoss is used.
        :type criterion: Optional[Union[Dict[str, Union[torch.nn.Module, Callable]], torch.nn.Module, Callable]]
        :param kwargs: The keyword arguments to pass to the BaseCallback.

        :keyword bool save_state: Whether to save the state of the optimizer. Defaults to True.
        :keyword bool load_state: Whether to load the state of the optimizer. Defaults to True.
        :keyword bool maximize: Whether to maximize the loss. Defaults to False.
        """
        kwargs.setdefault("save_state", True)
        kwargs.setdefault("load_state", True)
        kwargs.setdefault("criterion", torch.nn.MSELoss())
        super().__init__(params=params, **kwargs)
        if params is None:
            params = []
        else:
            params = list(params)
        if layers is not None:
            if isinstance(layers, torch.nn.Module):
                layers = [layers]
            params.extend([param for layer in layers for param in layer.parameters() if param not in params])
        self.params: List[torch.nn.Parameter] = params
        self.layers = layers
        self._default_params_lr = kwargs.get("params_lr", 2e-4)
        self._default_weight_decay = kwargs.get("weight_decay", 1e-2)
        self.DEFAULT_OPTIMIZER_CLS = kwargs.get("default_optimizer_cls", self.DEFAULT_OPTIMIZER_CLS)
        self._default_optim_kwargs = kwargs.get(
            "default_optim_kwargs", {
                "weight_decay": self._default_weight_decay,
                "lr": self._default_params_lr,
                "maximize": kwargs.get("maximize", False),
            }
        )
        self.param_groups = []
        self.optimizer = optimizer
        self.criterion = criterion

    def load_checkpoint_state(self, trainer, checkpoint: dict, **kwargs):
        if self.save_state:
            state = checkpoint.get(self.name, {})
            opt_state_dict = state.get(self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY, None)
            if opt_state_dict is not None:
                self.optimizer.load_state_dict(opt_state_dict)

    def get_checkpoint_state(self, trainer, **kwargs) -> object:
        if self.save_state:
            if self.optimizer is not None:
                return {
                    self.CHECKPOINT_OPTIMIZER_STATE_DICT_KEY: self.optimizer.state_dict()
                }
        return None

    def initialize_param_groups(self):
        """
        The learning rate are initialize. If the user has provided a learning rate for each parameter, then it is used.

        :return:
        """
        self.param_groups = []
        list_insert_replace_at(
            self.param_groups,
            self.OPTIMIZER_PARAMS_GROUP_IDX,
            {"params": self.params, "lr": self._default_params_lr}
        )
        return self.param_groups

    def create_default_optimizer(self):
        """
        Create the default optimizer.

        :return: The optimizer to use for training.
        """
        if not self.param_groups:
            self.initialize_param_groups()
        self.optimizer = self.DEFAULT_OPTIMIZER_CLS(self.param_groups, **self._default_optim_kwargs)
        return self.optimizer

    def start(self, trainer, **kwargs):
        super().start(trainer)
        if self.params and self.optimizer:
            pass
        elif self.params and self.optimizer is None:
            self.optimizer = self.create_default_optimizer()
        elif not self.params and self.optimizer is not None:
            self.param_groups = self.optimizer.param_groups
            self.params.extend([
                param
                for i in range(len(self.optimizer.param_groups))
                for param in self.optimizer.param_groups[i]["params"]
            ])
        else:
            self.params = list(trainer.model.parameters())
            self.optimizer = self.create_default_optimizer()

        if self.criterion is None and getattr(trainer, "criterion", None) is not None:
            self.criterion = trainer.criterion

    def apply_criterion(self, pred_batch, y_batch, **kwargs):
        criterion = kwargs.get("criterion", self.criterion)
        if criterion is None:
            if isinstance(y_batch, dict):
                criterion = {key: torch.nn.MSELoss() for key in y_batch}
            else:
                criterion = torch.nn.MSELoss()

        if isinstance(criterion, dict):
            if isinstance(y_batch, torch.Tensor):
                y_batch = {k: y_batch for k in criterion}
            if isinstance(pred_batch, torch.Tensor):
                pred_batch = {k: pred_batch for k in criterion}
            assert isinstance(pred_batch, dict) and isinstance(y_batch, dict), \
                "If criterion is a dict, pred, y_batch and pred must be a dict too."
            batch_loss = sum([
                criterion[k](pred_batch[k], y_batch[k].to(pred_batch[k].device))
                for k in criterion
            ])
        else:
            if isinstance(pred_batch, dict) and len(pred_batch) == 1:
                pred_batch = pred_batch[list(pred_batch.keys())[0]]
            batch_loss = criterion(pred_batch, y_batch.to(pred_batch.device))
        return batch_loss

    def _make_optim_step(self, pred_batch, y_batch, retain_graph=False):
        self.optimizer.zero_grad()
        batch_loss = self.apply_criterion(pred_batch, y_batch)
        batch_loss.backward(retain_graph=retain_graph)
        self.optimizer.step()
        return batch_loss.detach_()

    def on_optimization_begin(self, trainer, **kwargs):
        y_batch = trainer.current_training_state.y_batch
        pred_batch = format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
        batch_loss = self._make_optim_step(pred_batch, y_batch)
        trainer.update_state_(batch_loss=batch_loss)

    def on_optimization_end(self, trainer, **kwargs):
        self.optimizer.zero_grad()

    def on_validation_batch_begin(self, trainer, **kwargs):
        y_batch = trainer.current_training_state.y_batch
        pred_batch = format_pred_batch(trainer.current_training_state.pred_batch, y_batch)
        batch_loss = self.apply_criterion(pred_batch, y_batch)
        trainer.update_state_(batch_loss=batch_loss)

    def extra_repr(self) -> str:
        return f"optimizer={self.optimizer}, criterion={self.criterion}"

